################################################################################
#
# Copyright 2025 ByteDance Ltd. and/or its affiliates. All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
################################################################################

import argparse
import datetime
import logging
import os
import re
import shutil
import subprocess
import sys
from functools import lru_cache
from pathlib import Path
from typing import Dict, Optional, Tuple

import torch

"""
CREATE TABLE flux_perf(
  id BIGINT UNSIGNED NOT NULL auto_increment PRIMARY KEY COMMENT 'the key',
  op_name VARCHAR(32) not null COMMENT 'such as "gemm_rs"/"ag_gemm"',
  exp_name VARCHAR(64) not null COMMENT 'such as "flux #rank0", "torch #rank0"',
  url VARCHAR(256) not null COMMENT 'such as "123456"',
  compile_args VARCHAR(1024) not null COMMENT 'nvshmem/ipc',
  hardware VARCHAR(64) not null COMMENT 'Nvidia_L20. from nvidia-smi',
  runtime_args VARCHAR(256) not null COMMENT 'such as "py311_cu124_th24"',
  shape VARCHAR(256) not null COMMENT 'such as "M_8192_N_8192_K_8192"',
  dtype VARCHAR(256) not null COMMENT 'such as "float16"',
  extra_args VARCHAR(1024) not null COMMENT 'from parse_args + world_size from env',
  timestamp DATETIME not null COMMENT 'perf timestamp',
  duration numeric(8, 4) not null COMMENT 'the duration',
  commit_args JSON not null COMMENT 'code branch, commit id, author or so',
  compile_args_detail JSON not null COMMENT 'with nvshmem or not',
  runtime_args_detail JSON not null COMMENT 'such as container',
  environment_variables JSON not null COMMENT 'such as CUDA_MAX_CONNECTIONS',
  extra_args_detail JSON not null COMMENT 'from parse_args + world_size from env'
)
ENGINE=InnoDB DEFAULT CHARSET=utf8 comment 'perf database for FLUX'
"""

_DB_ENGINE = None
_DB_PSM = "toutiao.mysql.flux_ci_perf_staging_write"
_DB_TABLE_BYTED = "flux_perf"  #  Don't pollute this RDS table. only for gitlab CI.
_DB_TABLE_GITHUB = "flux_perf_github"  # Don't pollute this RDS table. only for github CI.
_DB_TABLE_TEST = "flux_perf_test"  # flux_perf_test: Do anything you like to this
_DB_TABLE = os.getenv("FLUX_PERF_TABLE", _DB_TABLE_TEST)
_OP_NAME = None
_SHAPE_ARGS, _DTYPE_ARGS, _EXTRA_ARGS, _FULL_ARGS = None, None, None, None


def _sort_dict(dict_):
    return dict(sorted(dict_.items()))


def _get_world_size():
    return int(os.getenv("WORLD_SIZE", 1))


def set_global_args(op_name, args):
    "to move shape_args and dtype_args from log_perf interface and save some dup code"
    global _OP_NAME, _SHAPE_ARGS, _DTYPE_ARGS, _EXTRA_ARGS, _FULL_ARGS
    assert (
        _OP_NAME is None
        and _SHAPE_ARGS is None
        and _DTYPE_ARGS is None
        and _EXTRA_ARGS is None
        and _FULL_ARGS is None
    ), "set_global_args() can only be called once"
    SHAPE_FIELDS = [
        "M",
        "N",
        "K",  # for dense & Moe Gather+RS
        "B",
        "S",
        "H",
        "ffn_hidden_size",  # for Moe AG+Scatter
        "G",  # num experts
        "topk",  # topk
        "E",  # EP size
        "T",  # TP size
        "input_groups",  # for MOE Gather+RS
        "weight_groups",  # for MOE AG+Scatter
    ]
    DTYPE_FIELDS = ["dtype"]
    FILTER_FIELDS = [
        "profile",
        "dist",
        "iters",
        "warmup",
        "triton",
        "lego",
        "triton_ony",
        "debug",
        "tune",
    ]  # those fields should never affect the performance
    args = args.__dict__
    _SHAPE_ARGS = _sort_dict({k: v for k, v in args.items() if k in SHAPE_FIELDS})
    _DTYPE_ARGS = _sort_dict({k: v for k, v in args.items() if k in DTYPE_FIELDS})
    _EXTRA_ARGS = _sort_dict(
        {k: v for k, v in args.items() if k not in DTYPE_FIELDS + SHAPE_FIELDS + FILTER_FIELDS}
    )
    _FULL_ARGS = _sort_dict(dict(args.items()))
    _EXTRA_ARGS.update({"TP": _get_world_size()})
    _OP_NAME = op_name
    assert _OP_NAME in ["gemm_rs", "ag_gemm", "moe_gather_rs", "moe_ag_scatter"]


def _db_engine():
    global _DB_ENGINE
    from bytedmysql import sqlalchemy_init
    from bytedmysql.sqlalchemy.dialect import BytedMySQLDialect
    from sqlalchemy import create_engine

    BytedMySQLDialect.supports_statement_cache = True

    if _DB_ENGINE is None:
        sqlalchemy_init()
        _DB_ENGINE = create_engine(f"mysql+pymysql://:@/?charset=utf8mb4&&db_psm={_DB_PSM}")
    return _DB_ENGINE


def should_log_to_rds():
    if os.getenv("CUDA_LAUNCH_BLOCKING", "0") == "1":
        logging.warning(f"with CUDA_LAUNCH_BLOCKING=1, no data with be logged to RDS")
    return os.getenv("FLUX_LOG_TO_RDS", "0") == "1"


def _commit_args() -> Dict[str, str]:
    return {
        "CI_ACTOR": os.getenv("CI_ACTOR"),
        "CI_HEAD_BRANCH": os.getenv("CI_HEAD_BRANCH"),
        "CI_HEAD_SHA": os.getenv("CI_HEAD_SHA"),
        "CI_JOB_ID": os.getenv("CI_JOB_ID"),
        "CI_JOB_IMAGE": os.getenv("CI_JOB_IMAGE"),
        "CI_JOB_NAME": os.getenv("CI_JOB_NAME"),
        "CI_PIPELINE_NAME": os.getenv("CI_PIPELINE_NAME"),
        "CI_PIPELINE_RUN_ID": os.getenv("CI_PIPELINE_RUN_ID"),
        "CI_REPO_NAME": os.getenv("CI_REPO_NAME"),
        "CI_REPO_WORKSPACE": os.getenv("CI_REPO_WORKSPACE"),
        "CI_STEP_ID": os.getenv("CI_STEP_ID"),
        "CI_STEP_NAME": os.getenv("CI_STEP_NAME"),
        "CI_EVENT_CHANGE_URL": os.getenv("CI_EVENT_CHANGE_URL", ""),
    }


def _cuda_version() -> Tuple[int, ...]:
    """CUDA Toolkit version as a (major, minor) by nvcc --version"""

    # Try finding NVCC
    nvcc_bin: Optional[Path] = None
    if nvcc_bin is None and os.getenv("CUDA_HOME"):
        # Check in CUDA_HOME
        cuda_home = Path(os.getenv("CUDA_HOME"))
        nvcc_bin = cuda_home / "bin" / "nvcc"
    if nvcc_bin is None:
        # Check if nvcc is in path
        nvcc_bin = shutil.which("nvcc")
        if nvcc_bin is not None:
            nvcc_bin = Path(nvcc_bin)
    if nvcc_bin is None:
        # Last-ditch guess in /usr/local/cuda
        cuda_home = Path("/usr/local/cuda")
        nvcc_bin = cuda_home / "bin" / "nvcc"
    if not nvcc_bin.is_file():
        raise FileNotFoundError(f"Could not find NVCC at {nvcc_bin}")

    # Query NVCC for version info
    output = subprocess.run(
        [nvcc_bin, "-V"],
        capture_output=True,
        check=True,
        universal_newlines=True,
    )
    match = re.search(r"release\s*([\d.]+)", output.stdout)
    version = match.group(1).split(".")
    return tuple(int(v) for v in version)


def _hardware() -> str:
    return torch.cuda.get_device_name().replace(" ", "_").lower()


def _is_perf_related_envs(key):
    return (key.startswith("CUDA_") and key not in ["CUDA_HOME", "CUDA_PATH"]) or (
        key.startswith("FLUX_")
        and key not in ["FLUX_LOG_TO_RDS", "FLUX_PERF_TABLE", "FLUX_EXTRA_TORCHRUN_ARGS"]
    )


@lru_cache
def _environment_variables():
    envs = _sort_dict(
        {key: value for key, value in os.environ.items() if _is_perf_related_envs(key)}
    )
    return envs


@lru_cache
def _runtime_args_detail():
    return {
        "torch": torch.__version__,
        "cuda": ".".join([str(x) for x in _cuda_version()]),
        "python": sys.version.split(" ")[0],  # return things like "3.11.2"
    }


@lru_cache
def _runtime_args():
    args = _runtime_args_detail()
    return "py{}_cu{}_th{}".format(
        "".join(args["python"].split(".")[:2]),
        "".join(args["cuda"].split(".")[:2]),
        "".join(args["torch"].split(".")[:2]),
    )


def _compile_with_nvshmem() -> bool:
    ci_job_name = os.getenv("CI_JOB_NAME", "")
    if ci_job_name.lower().find("nvshmem") >= 0:
        return True
    if ci_job_name.lower().find("ipc") >= 0:
        return False
    return True


def _compile_args_detail():
    return {
        "BUILD_WITH_NVSHMEM": _compile_with_nvshmem(),
    }


def _compile_args():
    if _compile_with_nvshmem():
        return "nvshmem"
    else:
        return "ipc"


def _extract_perf_result(perf_result) -> Tuple[str, float]:
    perf_result = perf_result.__dict__
    name = perf_result.get("name", "unknown")
    duration = perf_result.get("total_ms", 0.0)
    return name.replace(" ", ""), float(duration)


def _as_string(dict):
    def _prettify(v):
        if isinstance(v, bool):
            return "1" if v else "0"
        return str(v)

    return "_".join([f"{k}={_prettify(v)}" for k, v in dict.items()])


def _table_cls(table_name):
    from sqlalchemy import DATETIME, INT, JSON, NUMERIC, VARCHAR, Column
    from sqlalchemy.orm import declarative_base

    Base = declarative_base()

    # 定义User对象:
    class PerfItem(Base):
        # 表的名字:
        __tablename__ = _DB_TABLE

        id = Column(INT, primary_key=True)
        op_name = Column(VARCHAR(32))  # such as "gemm_rs"/"ag_gemm"
        exp_name = Column(VARCHAR(64))  # such as "flux #rank0", "torch #rank0"
        url = Column(VARCHAR(256))
        compile_args = Column(VARCHAR(1024))
        hardware = Column(VARCHAR(64))
        runtime_args = Column(VARCHAR(256))
        shape = Column(VARCHAR(256))
        dtype = Column(VARCHAR(256))
        extra_args = Column(VARCHAR(1024))
        timestamp = Column(DATETIME)
        duration = Column(NUMERIC(18, 2))
        # details fields: for API query
        commit_args = Column(JSON)
        compile_args_detail = Column(JSON)
        runtime_args_detail = Column(JSON)
        environment_variables = Column(JSON)
        extra_args_detail = Column(JSON)

        def __repr__(self) -> str:
            return (
                f"op_name: {self.op_name}, "
                f"exp_name: {self.exp_name}, "
                f"url: {self.url}, "
                f"compile_args: {self.compile_args}, "
                f"hardware: {self.hardware}, "
                f"runtime_args: {self.runtime_args}, "
                f"shape: {self.shape}, "
                f"dtype: {self.dtype}, "
                f"extra_args: {self.extra_args}, "
                f"timestamp: {self.timestamp}, "
                f"duration: {self.duration}, "
            )

    return PerfItem


def upload_to_rds(perf_result):
    from sqlalchemy.orm import sessionmaker

    PerfItem = _table_cls(_DB_TABLE)
    name, duration = _extract_perf_result(perf_result)
    item = PerfItem(
        op_name=_OP_NAME,
        exp_name=name,
        url=_commit_args()["CI_EVENT_CHANGE_URL"],
        compile_args=_compile_args(),
        hardware=_hardware(),
        runtime_args=_runtime_args(),
        shape=_as_string(_SHAPE_ARGS),
        dtype=_as_string(_DTYPE_ARGS),
        extra_args=_as_string(_EXTRA_ARGS),
        timestamp=datetime.datetime.now(),
        duration=duration,
        commit_args=_commit_args(),
        compile_args_detail=_compile_args_detail(),
        runtime_args_detail=_runtime_args_detail(),
        environment_variables=_environment_variables(),
        extra_args_detail=_FULL_ARGS,
    )
    logging.debug(item)

    engine = _db_engine()
    DBSession = sessionmaker(bind=engine)
    session = DBSession()
    session.add(item)
    session.commit()
    session.close()


def query(delete: bool = False):
    """sample if you want to query and delete some records"""
    from sqlalchemy.orm import sessionmaker

    PerfItem = _table_cls(_DB_TABLE)
    engine = _db_engine()
    DBSession = sessionmaker(bind=engine)
    session = DBSession()
    query = (
        session.query(PerfItem)
        .filter(PerfItem.environment_variables["CUDA_LAUNCH_BLOCKING"].as_string() == "1")
        .filter(PerfItem.duration == 3.0)
    )
    session.commit()
    for out in query.all():
        print(out)
    if delete:
        query.delete()
        session.commit()
    session.close()


def log_perf_stdout(perf_result):
    print(perf_result)


def log_perf_rds(perf_result):
    logging.debug(f"log to RDS")
    upload_to_rds(perf_result)


def log_perf(perf_result):
    log_perf_stdout(perf_result)
    if should_log_to_rds():
        try:
            log_perf_rds(perf_result)
        except Exception as e:
            logging.exception(f"Failed to upload perf result to RDS: {e}")


class PerfResult:
    def __init__(
        self,
        name: str,
        output: torch.Tensor,
        gathered_output: torch.Tensor,
        total_ms: float,
        time1: str,
        gemm_time_ms: float,
        time2: str,
        comm_time_ms: float,
        time3: str = "gemm_only",
        gemm_only_time_ms: float = 0,
    ) -> None:
        self.name = name
        self.output = output
        self.gathered_output = gathered_output
        self.total_ms = total_ms
        self.time1 = time1
        self.time2 = time2
        self.gemm_time_ms = gemm_time_ms
        self.comm_time_ms = comm_time_ms
        self.time3 = time3
        self.gemm_only_time_ms = gemm_only_time_ms

    def __repr__(self) -> str:
        if self.gemm_only_time_ms == 0.0:
            return (
                f"{self.name}: total {self.total_ms:.3f} ms, {self.time1} {self.gemm_time_ms:.3f} ms"
                f", {self.time2} {self.comm_time_ms:.3f} ms"
            )
        else:
            return (
                f"{self.name}: total {self.total_ms:.3f} ms, {self.time1} {self.gemm_time_ms:.3f} ms"
                f", {self.time2} {self.comm_time_ms:.3f} ms, {self.time3} {self.gemm_only_time_ms:.3f} ms"
            )


def _mock_ag_gemm_args():
    args = argparse.Namespace()
    setattr(args, "dtype", "s8")
    setattr(args, "M", 512)
    setattr(args, "N", 1024)
    setattr(args, "K", 8192)
    setattr(args, "iters", 10)
    setattr(args, "warmup", 5)
    setattr(args, "iters", 100)
    setattr(args, "transpose_weight", True)
    setattr(args, "has_bias", True)
    setattr(args, "fastacc", False)
    setattr(args, "ring_mode", "auto")
    setattr(args, "verify", False)
    setattr(args, "use_cuda_core_local", True)
    setattr(args, "use_cuda_core_ag", True)
    setattr(args, "triton", True)
    setattr(args, "gather_input", True)
    setattr(args, "debug", False)
    setattr(args, "profile", False)
    return args


if __name__ == "__main__":

    def parse_args():
        parser = argparse.ArgumentParser()
        parser.add_argument("--op", type=str, choices=["query", "log_perf"], default="log_perf")
        parser.add_argument("--delete", action="store_true", default=False)
        return parser.parse_args()

    args = parse_args()

    if "query" == args.op:
        query(args.delete)
    elif args.op == "log_perf":
        set_global_args("ag_gemm", _mock_ag_gemm_args())
        perf_res = PerfResult("flux #rank 0", None, None, 3.0, "gemm", 0.0, "comm", 0.0)
        log_perf(perf_res)



################################################################################
#
# Copyright 2025 ByteDance Ltd. and/or its affiliates. All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
################################################################################
from typing import List, Union

import torch
import torch.distributed

import flux
from flux.testing import matmul_int8


def moe_gather_rs_forward_torch(
    TP_GROUP,
    M: int,
    eid_start: int,
    ep_rank_m_start: int,
    ep_rank_m_end: int,
    inputs: Union[torch.Tensor, List[torch.Tensor]],
    weights: Union[torch.Tensor, List[torch.Tensor]],
    split_cpu: torch.Tensor,
    token_index: torch.Tensor,
    topk_index: torch.Tensor,
    topk: int,
    input_scales: Union[torch.Tensor, List[torch.Tensor]],
    weight_scales: Union[torch.Tensor, List[torch.Tensor]],
    output_vec_scales: Union[torch.Tensor, List[torch.Tensor]],
    do_all_reduce: bool = False,
    fast_acc: bool = False,
):
    if isinstance(inputs, torch.Tensor):
        inputs = [inputs]
        weights = [weights]
        input_scales = [input_scales]
        weight_scales = [weight_scales]
        output_vec_scales = [output_vec_scales]

    input0 = inputs[0]
    input_dtype = input0.dtype
    _, N, _ = weights[0].shape
    is_fp8 = flux.util.is_fp8_dtype(input_dtype)
    is_s8 = input_dtype == torch.int8
    output_type = torch.bfloat16 if is_fp8 or is_s8 else input_dtype

    # convert to list first
    full_output = torch.zeros((M, N), dtype=output_type, device=input0.device)
    for input_gid in range(len(inputs)):
        input = inputs[input_gid]
        weight = weights[input_gid]
        input_scale = input_scales[input_gid]
        weight_scale = weight_scales[input_gid]
        output_vec_scale = output_vec_scales[input_gid]

        acc = 0
        output_list = []

        gemm_only_op = flux.GemmOnly(
            input.dtype, output_type, use_fp8_gemm=flux.util.get_arch() < 90 and is_fp8
        )

        for exp_id in range(weight.size(0)):
            exp_w = weight[exp_id]
            Mi = split_cpu[exp_id + eid_start]
            exp_input = input[acc : acc + Mi]
            if not is_s8:
                scale_v = weight_scale[exp_id].item() * input_scale[0].item()
            else:
                scale_v = weight_scale[exp_id, None, :] * input_scale[acc : acc + Mi, None]
            acc += Mi
            if is_fp8:
                output_buf = torch.empty(Mi, N, dtype=torch.bfloat16).to(input.device)
                gemm_only_op.forward(exp_input, exp_w, output_buf=output_buf, fast_accum=fast_acc)
                output_list.append(scale_v * output_buf)
            elif is_s8 and Mi > 0:
                output = matmul_int8(exp_input, exp_w.t()).to(torch.float32) * scale_v
                output_list.append(output)
            else:
                output_list.append(scale_v * torch.matmul(exp_input, exp_w.t()))
        # M N
        output = torch.concat(output_list)
        # print(output.size())
        # print(output_vec_scale.size())
        assert output.size(0) == output_vec_scale.size(0)
        output = (output * output_vec_scale.unsqueeze(1)).to(output_type)

        new_index = (
            topk * token_index[ep_rank_m_start:ep_rank_m_end]
            + topk_index[ep_rank_m_start:ep_rank_m_end]
        )

        output1 = torch.zeros_like(full_output)
        output1[new_index] = output
        full_output += output1
    topk_reduce = full_output.view((full_output.size(0) // topk, topk, full_output.size(1))).sum(1)
    if do_all_reduce:
        torch.distributed.all_reduce(
            topk_reduce,
            op=torch.distributed.ReduceOp.SUM,
            group=TP_GROUP,
        )
        output2 = topk_reduce
    else:
        output2 = torch.zeros(
            (full_output.size(0) // TP_GROUP.size() // topk, full_output.size(1)),
            dtype=topk_reduce.dtype,
            device=torch.cuda.current_device(),
            requires_grad=False,
        )
        torch.distributed.reduce_scatter_tensor(output2, topk_reduce, group=TP_GROUP)
    return output2



################################################################################
#
# Copyright 2025 ByteDance Ltd. and/or its affiliates. All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
################################################################################

import gzip
import json
import logging
import pathlib
import shutil
import sys
from contextlib import contextmanager, nullcontext
from typing import List, Optional

import torch
import torch.distributed


def get_arch():
    properties = torch.cuda.get_device_properties(torch.cuda.current_device())
    major = properties.major
    minor = properties.minor
    return major * 10 + minor


def torch_allclose(x, y, rtol, atol, verbose=True):
    if not torch.allclose(x, y, rtol=rtol, atol=atol):
        print(f"shape of x: {x.shape}")
        print(f"shape of y: {y.shape}")

        print("x:", file=sys.stderr)
        print(x, file=sys.stderr)
        print("y:", file=sys.stderr)
        print(y, file=sys.stderr)
        print("x-y", x - y, file=sys.stderr)
        diff_loc = torch.isclose(x, y, rtol=rtol, atol=atol) == False
        print("x diff:", file=sys.stderr)
        print(x[diff_loc], file=sys.stderr)
        print("y diff:", file=sys.stderr)
        print(y[diff_loc], file=sys.stderr)
        num_diff = torch.sum(diff_loc)

        if len(y.shape) == 1:
            diff_rate = num_diff / y.shape[0]
        else:
            diff_rate = num_diff / (y.shape[0] * y.shape[1])
        print(f"diff count: {num_diff} ({diff_rate*100:.3f}%), {list(y.shape)}", file=sys.stderr)
        max_diff = torch.max(torch.abs(x - y))
        rtol_abs = rtol * torch.min(torch.abs(y))
        print(f"diff max: {max_diff}, atol: {atol}, rtol_abs: {rtol_abs}", file=sys.stderr)
        diff_indices = (diff_loc == True).nonzero(as_tuple=False)
        print(f"diff locations:\n{diff_indices}", file=sys.stderr)
        print("--------------------------------------------------------------\n", file=sys.stderr)
        raise RuntimeError

    if verbose:
        print("all close!")


def get_torch_prof_ctx(do_prof: bool):
    ctx = (
        torch.profiler.profile(
            activities=[torch.profiler.ProfilerActivity.CPU, torch.profiler.ProfilerActivity.CUDA],
            record_shapes=True,
            with_stack=False,
        )
        if do_prof
        else nullcontext()
    )
    return ctx


def _merge_json(to_merge_files: List[pathlib.Path], output_json: pathlib.Path):
    events = []
    for json_file in to_merge_files:
        if str(json_file).endswith("merged.json"):
            continue
        with open(json_file, "rb") as f:
            logging.debug(f"merge {json_file}")
            full_tl_json = json.loads(f.read().decode("latin-1"), strict=False)

        rank = full_tl_json["distributedInfo"]["rank"]
        world_size = full_tl_json["distributedInfo"]["world_size"]
        for e in full_tl_json["traceEvents"]:
            e["pid"] = f"{e['pid']}_{rank}"
            if isinstance(e["tid"], int):
                e["tid"] = e["tid"] * world_size + rank
            if e["name"] == "thread_name":
                e["args"]["name"] = f'{e["args"]["name"]}_{rank}'
            if e["name"] == "thread_sort_index":  # perfetto does not respect this.
                e["args"]["sort_index"] = e["args"]["sort_index"] * world_size + rank
        events.extend(full_tl_json["traceEvents"])

    with open(output_json, "w") as f:
        full_tl_json["traceEvents"] = events
        json.dump(events, f)


class group_profile:
    def __init__(
        self,
        name: str = None,
        do_prof: bool = True,
        merge_group: bool = True,
        keep_merged_only: bool = True,
        compress: bool = True,
        group: Optional[torch.distributed.ProcessGroup] = None,
    ):
        self.name = name
        self.do_prof = do_prof
        self.profile = torch.profiler.profile(
            activities=[torch.profiler.ProfilerActivity.CPU, torch.profiler.ProfilerActivity.CUDA],
            record_shapes=True,
            with_stack=True,
        )
        self.group = group or torch.distributed.group.WORLD
        self.merge_group = merge_group
        self.keep_merged_only = keep_merged_only
        self.compress = compress

    def __enter__(self):
        if self.do_prof:
            self.profile.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.do_prof:
            self.profile.__exit__(exc_type, exc_val, exc_tb)
            # export chrome trace
            outfile = pathlib.Path("prof") / f"{self.name}" / f"rank{self.group.rank()}.json"
            outfile.mkdir(parents=True, exist_ok=True)
            print(f"export chrome trace to {outfile}")
            self.profile.export_chrome_trace(str(outfile))
            if self.merge_group:
                self.merge_all()

    def merge_all(self):
        print(f"merge profiles...")
        # merge all
        if self.merge_group:
            torch.cuda.synchronize()  # wait for all ranks export
            torch.distributed.barrier(group=self.group)
            torch.cuda.synchronize()  # wait for all ranks export
        if self.group.rank() != 0:
            return

        # merge all json
        outdir = pathlib.Path("prof") / f"{self.name}"
        to_merge_files = outdir.glob("*.json")
        merged_json = pathlib.Path("prof") / f"{self.name}_merged.json"
        _merge_json(to_merge_files, merged_json)
        print(f"merge all profiles into {merged_json}")
        # here is an issue with python 3.10+ & with_stack=True save gz failure: https://github.com/pytorch/pytorch/issues/113564
        if self.compress:
            with open(merged_json, "rb") as f:
                with gzip.open(merged_json.with_suffix(".json.gz"), "wb") as g:
                    g.write(f.read())
            merged_json.unlink()
        if self.keep_merged_only:
            logging.info(f"remove profile directory: {outdir}")
            shutil.rmtree(outdir)


def is_fp8_dtype(dtype: torch.dtype) -> bool:
    return dtype.itemsize == 1 and dtype.is_floating_point


@contextmanager
def with_torch_deterministic(mode: bool, warn_only: bool = True):
    old_mode = torch.are_deterministic_algorithms_enabled()
    torch.use_deterministic_algorithms(mode, warn_only=warn_only)
    try:
        yield
    finally:
        torch.use_deterministic_algorithms(old_mode, warn_only=warn_only)


def bench_func(func, iters, warmup_iters):
    start_event = torch.cuda.Event(enable_timing=True)
    stop_event = torch.cuda.Event(enable_timing=True)
    for n in range(iters + warmup_iters):
        if n == warmup_iters:
            start_event.record()
        output = func()
    stop_event.record()
    start_event.wait()
    stop_event.wait()
    torch.cuda.current_stream().synchronize()
    duration_ms = start_event.elapsed_time(stop_event)
    return output, duration_ms / iters


__all__ = [
    "is_fp8_dtype",
    "get_arch",
    "torch_allclose",
    "get_torch_prof_ctx",
    "with_torch_deterministic",
    "bench_func",
    "group_profile",
]



################################################################################
#
# Copyright 2025 ByteDance Ltd. and/or its affiliates. All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
################################################################################

import datetime
import os
import random
from typing import Callable, List, Tuple, Union, Sequence

import numpy as np
import torch
import torch.distributed

import flux

_TP_LOCAL_GROUP = None
_TP_GROUP = None

DTYPE_MAP = {
    "bfloat16": torch.bfloat16,
    "float16": torch.float16,
    "float8_e4m3fn": torch.float8_e4m3fn,
    "float8_e5m2": torch.float8_e5m2,
    "s8": torch.int8,
    "s32": torch.int32,
}

RING_MODE_MAP = {
    "auto": None,
    "all2all": flux.AGRingMode.All2All,
    "ring1d": flux.AGRingMode.Ring1D,
    "ring2d": flux.AGRingMode.Ring2D,
}


def init_seed(seed=0):
    os.environ["NCCL_DEBUG"] = os.getenv("NCCL_DEBUG", "ERROR")
    os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":16:8"
    torch.use_deterministic_algorithms(True, warn_only=True)
    torch.set_printoptions(precision=2)
    torch.manual_seed(3 + seed)
    torch.cuda.manual_seed_all(3 + seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    torch.backends.cuda.matmul.allow_tf32 = False
    torch.backends.cuda.matmul.allow_fp16_reduced_precision_reduction = False
    torch.backends.cuda.matmul.allow_bf16_reduced_precision_reduction = False
    np.random.seed(3 + seed)
    random.seed(3 + seed)


def NNODES() -> int:
    WORLD_SIZE = int(os.environ.get("WORLD_SIZE", 1))
    LOCAL_WORLD_SIZE = int(os.environ.get("LOCAL_WORLD_SIZE", 1))
    return WORLD_SIZE // LOCAL_WORLD_SIZE


def initialize_distributed():
    global _TP_GROUP
    assert _TP_GROUP is None, "TP_GROUP has already been initialized"

    RANK = int(os.environ.get("RANK", 0))
    LOCAL_RANK = int(os.environ.get("LOCAL_RANK", 0))
    WORLD_SIZE = int(os.environ.get("WORLD_SIZE", 1))
    torch.cuda.set_device(LOCAL_RANK)
    torch.distributed.init_process_group(
        backend="nccl", world_size=WORLD_SIZE, rank=RANK, timeout=datetime.timedelta(seconds=1800)
    )
    assert torch.distributed.is_initialized()
    # use all ranks as tp group
    _TP_GROUP = torch.distributed.new_group(ranks=list(range(WORLD_SIZE)), backend="nccl")

    init_seed(seed=RANK)
    flux.init_flux_shm(_TP_GROUP)
    torch.cuda.synchronize()
    return _TP_GROUP


def TP_GROUP() -> torch.distributed.ProcessGroup:
    global _TP_GROUP
    assert _TP_GROUP is not None, "TP_GROUP has not been initialized"
    return _TP_GROUP


def is_local_tp_group_initialized():
    return _TP_LOCAL_GROUP is not None


def init_local_groups():
    global _TP_LOCAL_GROUP
    assert _TP_LOCAL_GROUP is None, "TP_LOCAL_GROUP has already been initialized"
    assert _TP_GROUP is not None, "TP_GROUP has not been initialized"
    nnodes = NNODES()
    RANK = _TP_GROUP.rank()
    LOCAL_RANK = int(os.environ.get("LOCAL_RANK", 0))
    LOCAL_WORLD_SIZE = int(os.environ.get("LOCAL_WORLD_SIZE", 1))
    if nnodes == 1:
        _TP_LOCAL_GROUP = _TP_GROUP
    else:
        for n in range(nnodes):
            ranks = list(range(LOCAL_WORLD_SIZE * n, LOCAL_WORLD_SIZE * (n + 1)))
            pg = torch.distributed.new_group(
                ranks=ranks,
                backend="nccl",
            )
            if RANK in ranks:
                _TP_LOCAL_GROUP = pg
    assert LOCAL_RANK == RANK % LOCAL_WORLD_SIZE
    assert _TP_LOCAL_GROUP.rank() == RANK % LOCAL_WORLD_SIZE
    assert _TP_LOCAL_GROUP.size() == LOCAL_WORLD_SIZE
    return _TP_LOCAL_GROUP


def _async_barrier():
    barrier_tensor = torch.tensor([1], device="cuda")
    torch.distributed.all_reduce(barrier_tensor, op=torch.distributed.ReduceOp.MAX, async_op=False)


def _make_tensor(
    shape: List[Union[int, Callable[[], int]]],
    dtype: torch.dtype,
    init_args: Union[Tuple[float, float], Tuple[int, int]],
    device: str = "cuda",
):
    """
    rand() * scale + bias
    randint(-scale, scale) + bias
    """
    if isinstance(shape, Sequence):
        shape = tuple([x() if isinstance(x, Callable) else x for x in shape])
    elif isinstance(shape, int):
        shape = (shape,)
    elif isinstance(shape, Callable):
        shape = shape()
    else:
        raise ValueError(f"unsupported shape {shape}")

    scale, bias = init_args
    if dtype in [torch.float16, torch.bfloat16, torch.float32]:
        out = (torch.rand(shape, dtype=dtype, device=device) * 2 - 1) * scale + bias
    elif dtype == torch.int8:
        out = torch.randint(-scale, scale, shape, dtype=torch.int8, device=device)
        out = out + bias
    elif flux.util.is_fp8_dtype(dtype):
        out = (torch.rand(shape, dtype=torch.float16, device=device) * 2 - 1) * scale + bias
        with flux.util.with_torch_deterministic(False):
            out = out.to(dtype)
    else:
        raise ValueError(f"unsupported dtype {dtype}")

    return out


def generate_data(configs):
    while True:
        yield (_make_tensor(*args) if args else None for args in configs)


def run_perf(expname, warmups, iters, func, sync_per_iter=False):
    start_events = [torch.cuda.Event(enable_timing=True) for _ in range(iters)]
    stop_events = [torch.cuda.Event(enable_timing=True) for _ in range(iters)]
    for n in range(warmups + iters):
        if n >= warmups:
            start_events[n - warmups].record()
            func(iter=n)
            stop_events[n - warmups].record()
        else:
            func(iter=n)
        if sync_per_iter:
            _async_barrier()
    [e.synchronize() for e in stop_events]
    elapsed_time_avg = (
        sum(
            [
                start_event.elapsed_time(stop_event)
                for start_event, stop_event in zip(start_events, stop_events)
            ]
        )
        / iters
    )
    print(f"exp: {expname}, avg: {elapsed_time_avg:0.3f} ms/iter")
    return elapsed_time_avg


def zeros_with_fp8(*args, **kwargs):
    with flux.util.with_torch_deterministic(False):
        return torch.zeros(*args, **kwargs)


def ones_with_fp8(*args, **kwargs):
    with flux.util.with_torch_deterministic(False):
        return torch.ones(*args, **kwargs)


def empty_non_determinisitc(*args, **kwargs):
    with flux.util.with_torch_deterministic(False):
        return torch.empty(*args, **kwargs)


def rand_with_fp8(*args, **kwargs):
    dtype = kwargs.get("dtype", None)
    kwargs.pop("dtype")
    if flux.util.is_fp8_dtype(dtype):
        with flux.util.with_torch_deterministic(False):
            return torch.rand(*args, dtype=torch.bfloat16, **kwargs).to(dtype)
    else:
        return torch.rand(*args, **kwargs)


def clone_with_fp8(tensor: torch.Tensor):
    with flux.util.with_torch_deterministic(False):
        return tensor.clone()


def all_gather_into_tensor_with_fp8(output_tensor, input_tensor, group=None, async_op=False):
    if flux.util.is_fp8_dtype(input_tensor.dtype):
        return torch.distributed.all_gather_into_tensor(
            output_tensor.view(torch.uint8),
            input_tensor.view(torch.uint8),
            group=group,
            async_op=async_op,
        )
    return torch.distributed.all_gather_into_tensor(
        output_tensor, input_tensor, group=group, async_op=async_op
    )


def matmul_int8(a, b):
    """
    torch._int_mm requires A.size(0) needs to be greater than 16
    """
    M, _ = a.shape
    if M <= 16:
        return torch._int_mm(torch.nn.functional.pad(a, (0, 0, 0, 32 - M)), b)[:M, :]
    return torch._int_mm(a, b)


def bitwise_eq(x: torch.Tensor, y: torch.Tensor):
    if x.shape != y.shape:
        return False
    if x.dtype != y.dtype:
        return False

    return torch.equal(x.view(torch.int8), y.view(torch.int8))


__all__ = [
    "DTYPE_MAP",
    "RING_MODE_MAP",
    "TP_GROUP",
    "NNODES",
    "init_seed",
    "initialize_distributed",
    "is_local_tp_group_initialized",
    "init_local_groups",
    "_async_barrier",
    "_make_tensor",
    "generate_data",
    "run_perf",
    "zeros_with_fp8",
    "ones_with_fp8",
    "rand_with_fp8",
    "empty_non_determinisitc",
    "all_gather_into_tensor_with_fp8",
    "matmul_int8",
    "bitwise_eq",
    "clone_with_fp8",
]



################################################################################
#
# Copyright 2025 ByteDance Ltd. and/or its affiliates. All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
################################################################################

from functools import partial

import torch
import torch.distributed as dist

import flux

FLUX_GEMM_RS_INTER_NODE_GROUP = None
INTRA_NODE_TP_GROUP = None

print = partial(print, flush=True)


def get_intra_node_pg_group(tp_group: torch.distributed.ProcessGroup, nnodes: int):
    global INTRA_NODE_TP_GROUP
    if INTRA_NODE_TP_GROUP is not None:
        return INTRA_NODE_TP_GROUP

    tp_world_size: int = tp_group.size()
    assert tp_world_size % nnodes == 0, f"{tp_world_size} is not divisible by {nnodes}"
    local_world_size: int = tp_group.size() // nnodes
    world_size: int = torch.distributed.get_world_size()
    assert world_size % tp_world_size == 0, f"{world_size} not divisible by {tp_world_size}"

    ranks_per_sub_group = []
    for node_id in range(nnodes):
        ranks_per_sub_group.append(
            list(range(node_id * local_world_size, (node_id + 1) * local_world_size))
        )

    INTRA_NODE_TP_GROUP, _ = torch.distributed.new_subgroups_by_enumeration(ranks_per_sub_group)
    print(f"split tp_group({tp_group.size()}) into {nnodes}x{INTRA_NODE_TP_GROUP.size()} subgroups")
    return INTRA_NODE_TP_GROUP


def get_inter_node_rs_group():
    global FLUX_GEMM_RS_INTER_NODE_GROUP
    assert (
        FLUX_GEMM_RS_INTER_NODE_GROUP is not None
    ), "FLUX_GEMM_RS_INTER_NODE_GROUP not initialized"
    return FLUX_GEMM_RS_INTER_NODE_GROUP


def initialize_inter_node_rs_group(tp_group: dist.ProcessGroup, nnodes: int):
    global FLUX_GEMM_RS_INTER_NODE_GROUP
    if FLUX_GEMM_RS_INTER_NODE_GROUP is not None:
        return

    tp_world_size: int = tp_group.size()
    assert tp_world_size % nnodes == 0, f"{tp_world_size} is not divisible by {nnodes}"
    local_world_size: int = tp_group.size() // nnodes
    world_size: int = dist.get_world_size()
    assert world_size % tp_world_size == 0, f"{world_size} not divisible by {tp_world_size}"
    cur_rank = dist.get_rank()

    for tp0_rank in range(0, world_size, tp_world_size):
        for local_rank in range(local_world_size):
            inter_ranks = [
                tp0_rank + node_rank * local_world_size + local_rank for node_rank in range(nnodes)
            ]
            inter_node_group = dist.new_group(inter_ranks)
            if cur_rank in inter_ranks:
                assert (
                    FLUX_GEMM_RS_INTER_NODE_GROUP is None
                ), "FLUX_GEMM_RS_INTER_NODE_GROUP already initialized"
                FLUX_GEMM_RS_INTER_NODE_GROUP = inter_node_group


class GemmRS_multinode:
    # use multi GemmRS and merge then from python
    def __init__(
        self,
        tp_group: dist.ProcessGroup,
        nnodes: int,
        max_m: int,
        n_dim: int,
        input_dtype: torch.dtype,
        transpose_weight: bool,
        fuse_reduction: bool,
    ) -> None:
        self.tp_group = tp_group
        self.rank = self.tp_group.rank()
        self.world_size = self.tp_group.size()
        self.nnodes = nnodes
        self.local_world_size = self.world_size // self.nnodes
        self.node_id = self.rank // self.local_world_size
        self.local_rank = self.rank % self.local_world_size
        self.max_m_total = max_m
        self.max_m = max_m // self.nnodes
        self.output_buffer = None
        self.cpp_op = None
        self.input_dtype = input_dtype
        self.transpose_weight = transpose_weight
        self.n_dim = n_dim
        self.fuse_reduction = fuse_reduction
        if self.nnodes > 1:
            self.tp_group_intra = get_intra_node_pg_group(tp_group, nnodes)
        else:
            self.tp_group_intra = tp_group
        self.lazy_init_cpp_op(self.max_m)

    def lazy_init_cpp_op(self, m_dim: int):
        if (self.cpp_op is not None) and m_dim <= self.max_m:
            return
        torch.distributed.barrier(self.tp_group_intra)
        self.max_m = m_dim
        self.cpp_op = flux.GemmRS(
            self.tp_group_intra,
            1,
            self.max_m,
            self.n_dim,
            self.input_dtype,
            transpose_weight=self.transpose_weight,
            fuse_reduction=self.fuse_reduction,
        )
        torch.distributed.barrier(self.tp_group_intra)
        torch.cuda.current_stream().synchronize()

    def forward(
        self, input: torch.Tensor, weight: torch.Tensor, bias: torch.Tensor
    ) -> torch.Tensor:
        stream = torch.cuda.current_stream()
        m_dim = input.size(0)
        assert m_dim % self.world_size == 0, f"{m_dim} % {self.world_size} != 0"
        self.lazy_init_cpp_op(m_dim // self.nnodes)
        outputs = []
        for n in range(self.nnodes):
            input_current = input[n * self.max_m : (n + 1) * self.max_m, ...]
            bias_current = bias[n * self.max_m : (n + 1) * self.max_m, ...] if bias else None
            # print(f"gemm rs node {n} with input: {input_current.shape} bias: {bias_current}")
            output = self.cpp_op.forward(input_current, weight, bias_current)
            output_buffer = torch.empty_like(output)
            output_buffer.copy_(output)
            outputs.append(output_buffer)
        # reduce buffer
        recv_buffer = torch.empty_like(outputs[0])
        output = outputs[self.node_id]

        for n in range(1, self.nnodes):
            node_send = (n + self.node_id) % self.nnodes
            rank_send = (n * self.local_world_size + self.rank) % self.world_size
            rank_recv = (self.rank - n * self.local_world_size + self.world_size) % self.world_size
            # print(f"n: {n} node_send: {node_send} rank_send: {rank_send} rank_recv: {rank_recv}")
            reqs = dist.batch_isend_irecv(
                [
                    dist.P2POp(dist.isend, outputs[node_send], rank_send, self.tp_group),
                    dist.P2POp(dist.irecv, recv_buffer, rank_recv, self.tp_group),
                ]
            )
            [req.wait() for req in reqs]
            output.add_(recv_buffer)

        return output



################################################################################
#
# Copyright 2025 ByteDance Ltd. and/or its affiliates. All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
################################################################################

import ctypes
from enum import Enum
from typing import List, Optional, Tuple
import typing

import torch
import torch.distributed as dist
import torch.distributed

def bsr_reduce(input: torch.Tensor, output: torch.Tensor, block_h: int, block_w: int): ...
def bitwise_check(A: torch.Tensor, B: torch.Tensor) -> bool: ...
def uniform_initialize(tensor: torch.Tensor, seed: int, min: float, max: float) -> bool: ...
def init_flux_shm(pg: dist.ProcessGroup) -> None: ...
def create_tensor(shape: List[int], dtype: torch.dtype, pg: dist.ProcessGroup) -> torch.Tensor: ...
def create_tensor_list(
    shape: List[int], dtype: torch.dtype, pg: dist.ProcessGroup
) -> List[torch.Tensor]: ...
def flux_create_shm_tensor_list(
    shape: List[int], dtype: torch.dtype, pg: dist.ProcessGroup
) -> torch.Tensor: ...
def topk_scatter_reduce(
    inputs: List[torch.Tensor], scatter_index: torch.Tensor, topk: int
) -> torch.Tensor: ...

class SegmentInfo:
    @property
    def segment_origin(self) -> int: ...
    @property
    def size(self) -> int: ...
    @property
    def tile_m_start_new(self) -> int: ...
    @property
    def tile_m_start_origin(self) -> int: ...

def get_gemm_rs_threadblock_segments_info(
    problem_shape: List[int],
    tiled_shape: List[int],
    rank: int,
    world_size: int,
    sub_world_size: int,
    nnodes: int,
    use_2d_ring: bool,
    per_tile_flags: bool,
) -> list[SegmentInfo]: ...
def calc_gemm_rs_threadblock_segments_info(
    segments: torch.Tensor,
    problem_shape: List[int],
    tiled_shape: List[int],
    rank: int,
    world_size: int,
    sub_world_size: int,
    nnodes: int,
    use_2d_ring: bool,
    per_tile_flags: bool,
): ...

class GroupBarrier:
    def __init__(self, pg: torch.distributed.ProcessGroup, ring_mode: bool) -> None:
        """
        group_name: torch.distributed.ProcessGroup.group_name
        ring_mode: use ring_mode barrier or not. force use ring_mode=True if torch.cuda.device_count() > 8
            use NVSHMEM implementation if compiled with --nvshmem and ring_mode=False, use flux implementation if not.
        """

    def barrier_all(self, stream: int) -> None: ...

class TuningRecord:
    pass

class AGRingMode:
    """
    Members:

      All2All

      Ring1D

      Ring2D
    """

    All2All: typing.ClassVar[AGRingMode]  # value = <AGRingMode.All2All: 0>
    Ring1D: typing.ClassVar[AGRingMode]  # value = <AGRingMode.Ring1D: 1>
    Ring2D: typing.ClassVar[AGRingMode]  # value = <AGRingMode.Ring2D: 2>
    __members__: typing.ClassVar[
        dict[str, AGRingMode]
    ]  # value = {'All2All': <AGRingMode.All2All: 0>, 'Ring1D': <AGRingMode.Ring1D: 1>, 'Ring2D': <AGRingMode.Ring2D: 2>}


class AllGatherOption:
    """
    TODO(houqi.1993)
        * for small input, use CUDA core implementation.

    input_buffer_copied: if input and input_scale is ready. default to False. If True,
    use_cuda_core_local: default to True for INT8 input with input_scale. otherwise False
    use_cuda_core_ag: default to True for INT8 input with input_scale with small shape. otherwise False
    fuse_sync: fuse sync into copy function.
    use_read: read for pull mode, write for push mode. for A100 NVLink, read mode is better or faster. otherwise write mode is better.
    mode: use AlltoAll mode or ring mode(Ring1D or Ring2D)
    """

    def __init__(self) -> None: ...
    input_buffer_copied: Optional[bool]
    use_cuda_core_local: Optional[bool]
    use_cuda_core_ag: Optional[bool]
    fuse_sync: Optional[bool]
    use_read: Optional[bool]
    mode: Optional[AGRingMode]

class RingMode:
    All2All: typing.ClassVar[RingMode]  # value = <RingMode.All2All: 0>
    Ring1D: typing.ClassVar[RingMode]  # value = <RingMode.Ring1D: 1>
    Ring2D: typing.ClassVar[RingMode]  # value = <RingMode.Ring2D: 2>
    __members__: typing.ClassVar[
        dict[str, RingMode]
    ]  # value = {'All2All': <RingMode.All2All: 0>, 'Ring1D': <RingMode.Ring1D: 1>, 'Ring2D': <RingMode.Ring2D: 2>}

class ReduceScatterOption:
    use_barrier_queue: Optional[bool]
    use_1d_ring: Optional[bool]
    use_p2p_read: Optional[bool]
    use_cudaMemcpyAsync: Optional[bool]
    use_gemmk: Optional[bool]
    per_tile_flags: Optional[bool]
    n_split: Optional[int]
    num_blocks: Optional[int]
    ring_mode: Optional[RingMode]

def load_tuning_record(record: TuningRecord) -> None: ...

class ProfilingContext:
    def __init__(self, name: str): ...
    def get_code(self) -> str: ...
    def get_latest_prof_result(self) -> str: ...
    def get_all_prof_results(self) -> List[str]: ...
    def get_latest_record(self) -> TuningRecord: ...
    def get_all_records(self) -> List[TuningRecord]: ...

class DistEnvTP:
    def __init__(tp_group: dist.ProcessGroup, nnodes: int) -> None: ...

class DistEnvTPWithEP:
    def __init__(
        tp_group: dist.ProcessGroup, nnodes: int, ep_group: Optional[dist.ProcessGroup] = None
    ) -> None: ...

class MoeArguments:
    def __init__(
        max_ntokens: int,
        hidden: int,
        ffn_hidden: int,
        nexperts: int,
        topk: int,
        input_dtype=torch.dtype,
        output_dtype: Optional[torch.dtype] = None,
    ) -> None: ...

class GemmOnly:
    """
    support 4 mode: FP16/BF16 mode, FP8 mode, INT8 Dequant mode, INT8 (GEMM)Only mode

    About shapes:
    * input: [M, K] for all types
    * weight: [K, N] if transpose_weight, [N, K] if not transpose_weight.
    * bias: [1, N] for FP8 or INT8 Dequant, [M, N] for FP16/BF16/INT8(Only).
    * input_scale: always None for FP16/BF16/Int8 Only. [M, 1] for INT8 Dequant. [1] for FP8
    * weight_scale: always None for FP16/BF16/Int8 Only. [1, N] for INT8 Dequant. [1] for FP8
    * output_scale: always None for FP16/BF16/INT8 Dequant/INT8 Only. for FP8 ???

    for transpose_weight=True:
        for FP16:  output_FP16 = [input_FP16 * weight_FP16]_FP32.to(FP16) + bias_FP16
        for BF16:  output_BF16 = [input_BF16 * weight_BF16]_FP32.to(BF16) + bias_BF16
        for FP8:   output_BF16 = [[input_FP8 * weight_FP8]_FP32 * input_scale_FP32 * weight_scale_FP32]_FP32.to(BF16) + bias_FP16
        for INT8 Dequant:  output_BF16 = [[input_INT8 * weight_INT8]_INT32.to(FP32) * input_scale_FP32 * weight_scale_FP32]_FP32.to(BF16) + bias_BF16
        for INT8 Only: output_INT32 = [input_INT8 * weight_INT8]_INT32 + bias_INT32
    for transpose_weight=False, replace weight with weight.T
    """

    def __init__(
        self,
        input_dtype: torch.dtype,
        output_dtype: Optional[torch.dtype] = None,
        transpose_weight: bool = False,
        use_fp8_gemm: bool = False,
    ): ...
    def forward(
        self,
        input: torch.Tensor,
        weight: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
        output_buf: Optional[torch.Tensor] = None,
        input_scale: Optional[torch.Tensor] = None,
        weight_scale: Optional[torch.Tensor] = None,
        output_scale: Optional[torch.Tensor] = None,
        fast_accum: bool = False,
    ) -> torch.Tensor: ...
    def profiling(
        self,
        input: torch.Tensor,
        weight: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
        output_buf: Optional[torch.Tensor] = None,
        input_scale: Optional[torch.Tensor] = None,
        weight_scale: Optional[torch.Tensor] = None,
        output_scale: Optional[torch.Tensor] = None,
        fast_accum: bool = False,
        prof_ctx: Optional[ProfilingContext] = None,
    ) -> torch.Tensor: ...

class BlockScaleGemm:
    def __init__(
        self,
        input_dtype: torch.dtype,
        output_dtype: Optional[torch.dtype] = None,
    ): ...
    def forward(
        self,
        input: torch.Tensor,
        weight: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
        output: Optional[torch.Tensor] = None,
        input_scale: Optional[torch.Tensor] = None,
        weight_scale: Optional[torch.Tensor] = None,
    ) -> torch.Tensor: ...
    def forward_multi_stream(
        self,
        input: torch.Tensor,
        input_list: torch.Tensor,
        weight: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
        output: Optional[torch.Tensor] = None,
        input_scale: Optional[torch.Tensor] = None,
        weight_scale: Optional[torch.Tensor] = None,
    ) -> torch.Tensor: ...
    def reference(
        self,
        input: torch.Tensor,
        weight: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
        output: Optional[torch.Tensor] = None,
        input_scale: Optional[torch.Tensor] = None,
        weight_scale: Optional[torch.Tensor] = None,
    ) -> torch.Tensor: ...
    def profiling(
        self,
        input: torch.Tensor,
        weight: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
        output: Optional[torch.Tensor] = None,
        input_scale: Optional[torch.Tensor] = None,
        weight_scale: Optional[torch.Tensor] = None,
        prof_ctx: Optional[ProfilingContext] = None,
    ) -> torch.Tensor: ...

class GemmRS:
    def __init__(
        self,
        tp_group: dist.ProcessGroup,
        nnodes: int,
        max_m: int,
        n_dim: int,
        input_dtype: torch.dtype,
        output_dtype: Optional[torch.dtype] = None,
        transpose_weight=False,
        fuse_reduction=False,
        ring_reduction=False,
    ):
        """The formula:
        * matmul(input, weight) * input_scale * weight_scale + bias   if transpose_weight=True
        * matmul(input, weight.T) * input_scale * weight_scale + bias   if transpose_weight=False
        """

    def forward(
        self,
        input: torch.Tensor,
        weight: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
        input_scale: Optional[torch.Tensor] = None,
        weight_scale: Optional[torch.Tensor] = None,
        output_scale: Optional[torch.Tensor] = None,
        fast_accum: bool = False,
        reduce_scatter_option: ReduceScatterOption = ReduceScatterOption(),
    ) -> torch.Tensor:
        """
        support 3 mode: FP16/BF16 mode, FP8 mode, INT8 mode

        TODO(wenlei.bao)
        * what about output_scale?

        About shapes: M = M_per_rank * world_size
        * input: [M_per_rank, K] for all types
        * weight: [K, N] if transpose_weight, [N, K] if not transpose_weight.
        * bias: [1, N] for FP8 or INT8, [M_per_rank, N] for FP16 or BF16.
        * input_scale: always None for FP16/BF16. [M, 1] for INT8. [1] for FP8
        * weight_scale: always None for FP16/BF16. [1, N] for INT8. [1] for FP8
        * output_scale: always None for FP16/BF16/INT8. for FP8 ???

        for transpose_weight=True:
            for FP16:  output_FP16 = [input_FP16 * weight_FP16]_FP32.to(FP16) + bias_FP16
            for BF16:  output_BF16 = [input_BF16 * weight_BF16]_FP32.to(BF16) + bias_BF16
            for FP8:   output_BF16 = [[input_FP8 * weight_FP8]_FP32 * input_scale_FP32 * weight_scale_FP32]_FP32.to(BF16) + bias_FP16
            for INT8:  output_BF16 = [[input_INT8 * weight_INT8]_INT32.to(FP32) * input_scale_FP32 * weight_scale_FP32]_FP32.to(BF16) + bias_BF16
        for transpose_weight=False, replace weight with weight.T
        """

    def forward_barrier(
        self, input: torch.Tensor, weight: torch.Tensor, bias: Optional[torch.Tensor] = None
    ) -> None: ...
    def forward_reduce_scatter(
        self,
        input: torch.Tensor,
        weight: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
        input_scale: Optional[torch.Tensor] = None,
        weight_scale: Optional[torch.Tensor] = None,
    ) -> torch.Tensor: ...
    def zero_buffers(self) -> None: ...
    def profiling(
        self,
        input: torch.Tensor,
        weight: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
        input_scale: Optional[torch.Tensor] = None,
        weight_scale: Optional[torch.Tensor] = None,
        output_scale: Optional[torch.Tensor] = None,
        fast_accum: bool = False,
        prof_ctx: Optional[ProfilingContext] = None,
    ) -> torch.Tensor: ...

class GemmRSAcrossNode:
    def __init__(
        self,
        rank: int,
        world_size: int,
        local_world_size: int,
        weight: torch.Tensor,
        transpose_weight: bool,
        output_buffer: torch.Tensor,
        output: torch.Tensor,
        barrier: torch.Tensor,
        fuse_reduction: bool,
    ): ...
    def forward(self, input: torch.Tensor) -> torch.Tensor: ...

class AGKernel:
    def __init__(
        self,
        tp_group: dist.ProcessGroup,
        nnodes: int,
        full_m: int,
        n_dim: int,
        k_dim: int,
        input_dtype: torch.dtype,
        output_dtype: Optional[torch.dtype] = None,
    ): ...
    def forward(
        self,
        input: torch.Tensor,
        weight: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
        output: Optional[torch.Tensor] = None,
        input_scale: Optional[torch.Tensor] = None,
        weight_scale: Optional[torch.Tensor] = None,
        output_scale: Optional[torch.Tensor] = None,
        fast_accum: bool = False,
        transpose_weight: bool = False,
        all_gather_option: AllGatherOption = AllGatherOption(),
        gathered_input: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """
        support 3 mode: BF16/FP16 mode or FP8 mode or int8 mode

        TODO(wenlei.bao)
        * what about output_scale?

        About shapes: M = M_per_rank * world_size
        * input: [M_per_rank, K] for all types
        * weight: [K, N] if transpose_weight, [N, K] if not transpose_weight.
        * bias: [1, N] for FP8 or INT8, [M_per_rank, N] for FP16 or BF16.
        * output: [M, N]
        * input_scale: [M_per_rank, 1] for INT8, [1] for FP8.
        * weight_scale: [1, N] for INT8, [1] for FP8
        * output_scale: ?

        About Optional:
        * output: optional for all modes. if set, the output will be written to this buffer.
        * bias: optional for all modes. zero for None.
        * input_scale/weight_scale: always None for FP16/BF16. always not None for INT8. for FP8 ??
        * output_scale: always None for FP16/BF16/INT8. for FP8 ???
        * gathered_input: if not set to None, the input will be allgathered to this tensor.

        About dtypes and formula:
        * for FP16 mode:  output_FP16 = [input_FP16 * weight_FP16]_FP32.to(FP16) + bias_FP16
        * for BF16 mode:  output_BF16 = [input_BF16 * weight_BF16]_FP32.to(BF16) + bias_BF16
        * for FP8  mode:  output_BF16 = [[input_FP8 * weight_FP8]_FP32.to(BF16) * input_scale_FP32 * weight_scale_FP32]_FP32.to(BF16) + bias_BF16
        * for int8 mode:  output_BF16 = [[input_INT8 * weight_INT8]_INT32.to(FP32) * input_scale_FP32 * weight_scale_FP32]_FP32.to(BF16) + bias_BF16
        """

    def gemm_only(
        self,
        input: torch.Tensor,
        weight: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
        output: Optional[torch.Tensor] = None,
        input_scale: Optional[torch.Tensor] = None,
        weight_scale: Optional[torch.Tensor] = None,
        output_scale: Optional[torch.Tensor] = None,
        fast_accum: bool = False,
        transpose_weight: bool = False,
    ) -> torch.Tensor: ...
    def profiling(
        self,
        input: torch.Tensor,
        weight: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
        output: Optional[torch.Tensor] = None,
        input_scale: Optional[torch.Tensor] = None,
        weight_scale: Optional[torch.Tensor] = None,
        output_scale: Optional[torch.Tensor] = None,
        fast_accum: bool = False,
        transpose_weight: bool = False,
        all_gather_option: AllGatherOption = AllGatherOption(),
        gathered_input: Optional[torch.Tensor] = None,
        prof_ctx: Optional[ProfilingContext] = None,
    ) -> torch.Tensor: ...

class AGKernelCrossNode:
    def __init__(
        self,
        tp_group: dist.ProcessGroup,
        intra_node_group: torch.distributed.ProcessGroup,
        nnodes: int,
        output_buffer: torch.Tensor,
        full_m: int,
        n_dim: int,
        k_dim: int,
        input_dtype: torch.dtype,
        transpose_weight: bool = True,
        local_copy: bool = False,
        ring_mode: AGRingMode | None = None,
    ): ...
    def reset_signals(self) -> None: ...
    def copy_local(self, input: torch.Tensor) -> None: ...
    def forward(self, input: torch.Tensor, weight: torch.Tensor) -> torch.Tensor: ...
    def gemm_only(
        self, input: torch.Tensor, full_input: torch.Tensor, weight: torch.Tensor
    ) -> torch.Tensor: ...

class All2AllOp:
    def __init__(self, rank: int, world_size: int, recv_buffer: torch.Tensor): ...
    def forward(self, send_buffer: List[torch.Tensor]) -> None: ...

class GemmGroupedV2:
    def __init__(
        self, weight: torch.Tensor, num_experts: int, in_type: torch.dtype, out_type: torch.dtype
    ): ...
    def forward(
        self,
        input: torch.Tensor,
        split_cpu: torch.Tensor,
        input_scale: Optional[List[torch.Tensor]] = None,
        weight_scale: Optional[List[torch.Tensor]] = None,
        fast_accum: bool = False,
        sm_margin: int = 0,
    ) -> torch.Tensor: ...

class GemmGroupedV3:
    def __init__(self, weight: torch.Tensor, num_experts: int): ...
    def forward(self, input: torch.Tensor, split_cpu: torch.Tensor) -> torch.Tensor: ...
    def profiling(
        self, input: torch.Tensor, splits_cpu: torch.Tensor, prof_ctx: ProfilingContext = None
    ) -> torch.Tensor: ...

class GemmGroupedV3AGScatter:
    def __init__(self, tp_env: DistEnvTP, moe_args: MoeArguments): ...
    def forward(
        self,
        inputs_shard: torch.Tensor,
        weights: torch.Tensor,
        splits_gpu: torch.Tensor,
        scatter_index: torch.Tensor,
        output_scale: Optional[torch.Tensor] = None,
        outputs_buf: Optional[torch.Tensor] = None,
        allgather_output: Optional[torch.Tensor] = None,
        fast_accum: bool = False,
        sm_margin: int = 0,
    ) -> torch.Tensor: ...
    def forward_multiple_weights(
        self,
        inputs_shard: torch.Tensor,
        weights: List[torch.Tensor],
        splits_gpu: torch.Tensor,
        scatter_index: torch.Tensor,
        output_scale: Optional[List[torch.Tensor]] = None,
        outputs_buf: Optional[List[torch.Tensor]] = None,
        allgather_output: Optional[torch.Tensor] = None,
        fast_accum: bool = False,
        sm_margin: int = 0,
    ) -> List[torch.Tensor]: ...
    def clear_buffers(self) -> None: ...
    def profiling(
        self,
        inputs_shard: torch.Tensor,
        weights: List[torch.Tensor],
        splits_gpu: torch.Tensor,
        scatter_index: torch.Tensor,
        output_scale: Optional[List[torch.Tensor]] = None,
        outputs_buf: Optional[List[torch.Tensor]] = None,
        allgather_output: Optional[torch.Tensor] = None,
        fast_accum: bool = False,
        sm_margin: int = 0,
        prof_ctx: Optional[ProfilingContext] = None,
    ) -> List[torch.Tensor]: ...

class GemmGroupedV2AGScatterOp:
    def __init__(
        self,
        tp_env: DistEnvTP,
        moe_args: MoeArguments,
    ): ...
    def forward(
        self,
        inputs_shard: torch.Tensor,
        weights: torch.Tensor,
        splits_gpu: torch.Tensor,
        scatter_index: torch.Tensor,
        input_scale: Optional[torch.Tensor] = None,
        weight_scale: Optional[torch.Tensor] = None,
        output_scale: Optional[torch.Tensor] = None,
        outputs_buf: Optional[torch.Tensor] = None,
        allgather_output: Optional[torch.Tensor] = None,
        fast_accum: bool = False,
        sm_margin: int = 0,
        ag_option: AllGatherOption = AllGatherOption(),
    ) -> torch.Tensor: ...
    """
    2 modes supported: FP16/BF16, or FP8 mode.  INT8 is not supported yet.

    About some variables:
        * ntokens = tokens_per_rank * world_size.
        * M = ntokens * topk.
        * sum(M_this_ep) = M. for EP=1, M_this_ep = M.

    About shapes:
    * inputs_shard: [M_this_ep, K] for all types
    * weights: [E, K, N] if transpose_weight, [E, N, K] if not transpose_weight.
    * splits_gpu: [E] of torch.int32 by default or [E+1] if drop_token. for EP=1, M_this_rank = M * topk.
    * scatter_index: [ntokens, topk] of torch.int32
    * output_scale: [E] of torch.float32
    * output_buf: [M_this_ep, N].
    * allgather_output: [ntokens, K] of the same dtype as inputs_shard

    input_{T} = index_select(all_gather(inputs_shard_{T}), dim=0, index=gather_index)  for T in [FP16,BF16,FP8]
    for FP16 mode: output_FP16 = [input_FP16 * weights_FP16]_FP32.to(FP16)
    for BF16 mode: output_BF16 = [input_BF16 * weights_BF16]_FP32.to(BF16)
    for FP8  mode: output_BF16 = [[input_FP8 * weights_FP8]_FP32 * output_scale_FP32]_FP32.to(BF16)

    """

    def forward_multiple_weights(
        self,
        inputs_shard: torch.Tensor,
        weights: List[torch.Tensor],
        splits_gpu: torch.Tensor,
        scatter_index: torch.Tensor,
        input_scale: Optional[List[torch.Tensor]] = None,
        weight_scale: Optional[List[torch.Tensor]] = None,
        output_scale: Optional[List[torch.Tensor]] = None,
        outputs_buf: Optional[List[torch.Tensor]] = None,
        allgather_output: Optional[torch.Tensor] = None,
        fast_accum: bool = False,
        sm_margin: int = 0,
        ag_option: AllGatherOption = AllGatherOption(),
    ) -> List[torch.Tensor]: ...
    def clear_buffers(self) -> None: ...
    def profiling(
        self,
        inputs_shard: torch.Tensor,
        weights: List[torch.Tensor],
        splits_gpu: torch.Tensor,
        scatter_index: torch.Tensor,
        input_scale: Optional[List[torch.Tensor]] = None,
        weight_scale: Optional[List[torch.Tensor]] = None,
        output_scale: Optional[List[torch.Tensor]] = None,
        outputs_buf: Optional[List[torch.Tensor]] = None,
        allgather_output: Optional[torch.Tensor] = None,
        fast_accum: bool = False,
        sm_margin: int = 0,
        ag_option: AllGatherOption = AllGatherOption(),
        prof_ctx: Optional[ProfilingContext] = None,
    ) -> List[torch.Tensor]: ...

class GemmGroupedV3GatherRS:
    def __init__(
        self,
        num_experts: int,
        max_m: int,
        n_dim: int,
        topk: int,
        rank: int,
        world_size: int,
        tp_world_size: int,
        ep_world_size: int,
    ): ...
    def forward_gather_rs(
        self,
        input: torch.Tensor,
        weight: torch.Tensor,
        split_cpu: torch.Tensor,
        scatter_idx: torch.Tensor,
        input_scale: Optional[torch.Tensor] = None,
        weight_scale: Optional[torch.Tensor] = None,
        output_scale: Optional[torch.Tensor] = None,
        fastacc: bool = True,
        sm_margin: int = 0,
        with_stream_sync: bool = False,
    ) -> torch.Tensor: ...
    def forward_gather_rs_multiple(
        self,
        input: List[torch.Tensor],
        weight: List[torch.Tensor],
        split_cpu: torch.Tensor,
        scatter_idx: torch.Tensor,
        input_scale: Optional[List[torch.Tensor]] = None,
        weight_scale: Optional[List[torch.Tensor]] = None,
        output_scale: Optional[List[torch.Tensor]] = None,
        fastacc: bool = True,
        sm_margin: int = 0,
        with_stream_sync: bool = False,
    ) -> torch.Tensor: ...

class GemmGroupedV2GatherRSOp:
    def __init__(
        self,
        tp_group: dist.ProcessGroup,
        total_num_experts: int,
        max_m: int,
        n_dim: int,
        topk: int,
        output_dtype: torch.dtype,
        tp_world_size: int,
        ep_world_size: int,
        max_input_groups: int = 1,
        n_split: int = 4,
        do_all_reduce: bool = False,
        use_read_mode: bool = False,
    ):
        """
        such conditions expected:
            max_input_groups <= 2
            tp_world_size * ep_world_size == tp_group.size
        """
        ...

    def forward_gather_rs(
        self,
        input: torch.Tensor,
        weight: torch.Tensor,
        splits_cpu: torch.Tensor,
        scatter_idx: torch.Tensor,
        bias: Optional[torch.Tensor] = None,
        input_scale: Optional[torch.Tensor] = None,
        weight_scale: Optional[torch.Tensor] = None,
        output_vec_scale: Optional[torch.Tensor] = None,
        fast_accum: bool = True,
        sm_margin: int = 0,
        with_stream_sync: bool = False,  # NOTE: not used. to align with V3
    ) -> torch.Tensor:
        """
        support 3 modes: FP16/BF16 mode, or FP8(FP8E4M3FN/FP8E5M2) mode, or INT8 mode.

        TODO(houqi.1993) INT8 mode is not supported

        some params:
        * E_this_ep: sum(E_this_ep) = E. for EP=1, E_this_ep = E.
        * M_this_ep: sum(M_this_ep) = M. for EP=1, M_this_ep = M = ntokens * topk

        some shapes:
        * input: [M_this_ep, K] for all types
        * weight: [E_this_ep, K, N] if transpose_weight, [E_this_ep, N, K] if not transpose_weight.
        * bias: [N] for int8
        * input_scale: [1] for FP16/BF16/FP8, [M_this_ep] for INT8 dynamic quant
        * weight_scale: [E_this_ep, 1] for FP16/BF16/FP8, [E_this_ep, N] for INT8.

        for FP16:
            gemm_out_FP16(m_indexes) =
                [input(m_indexes(i), :)_FP16 * weight(i, :, :)_FP16]_FP32
                * input_scale(None, None)_FP32
                * weight_scale(i, None, None)_FP32
                * output_scale(m_indexes(i), None)_FP32]_FP32.to(FP16)  . for expert i in E_this_ep.
            scatter_idx -> gather_index
            output_FP16 = reduce_scatter(select_index(gemm_out_FP16, gather_index)_FP16)_FP16
        for BF16: replace FP16 with BF16 and use FP16 formula.
        for FP8(both FP8E4M3FN or FP8E5M2):
            gemm_out_BF16(m_indexes) =
                [input(m_indexes(i), :)_FP8 * weight(i, :, :)_FP8]_FP32
                * input_scale(None, None)_FP32
                * weight_scale(i, None, None)_FP32
                * output_scale(m_indexes(i), None)_FP32]_FP32.to(BF16)  . for expert i in E_this_ep.
            output_BF16 = reduce_scatter(select_index(gemm_out_BF16, gather_index)_BF16)_BF16
        for INT8:
            gemm_out_BF16(m_indexes) =
                [input(m_indexes(i), :)_INT8 * weight(i, :, :)_INT8]_FP32
                * weight_scale(i, None, :)_FP32
                * output_scale(m_indexes(i), None)_FP32]_FP32.to(BF16)
                 + bias . for expert i in E_this_ep.
            output_BF16 = reduce_scatter(select_index(gemm_out_BF16, gather_index)_BF16)_BF16
        """

    def forward_gather_rs_multiple(
        self,
        input: List[torch.Tensor],
        weight: List[torch.Tensor],
        splits_cpu: torch.Tensor,
        scatter_idx: torch.Tensor,
        input_scale: Optional[List[torch.Tensor]] = None,
        weight_scale: Optional[List[torch.Tensor]] = None,
        output_vec_scale: Optional[List[torch.Tensor]] = None,
        fast_accum: bool = True,
        sm_margin: int = 0,
        with_stream_sync: bool = False,
    ) -> torch.Tensor: ...
    def profiling(
        self,
        input: torch.Tensor,
        weight: torch.Tensor,
        splits_cpu: torch.Tensor,
        scatter_idx: torch.Tensor,
        input_scale: Optional[torch.Tensor] = None,
        weight_scale: Optional[torch.Tensor] = None,
        output_vec_scale: Optional[torch.Tensor] = None,
        fastacc: bool = True,
        sm_margin: int = 0,
        with_stream_sync: bool = False,
        prof_ctx: Optional[ProfilingContext] = None,
    ) -> torch.Tensor: ...

class TopkReduceScatterOp:
    def __init__(
        self,
        tp_group: dist.ProcessGroup,
        max_m: int,
        n_dim: int,
        topk: int,
        output_dtype: torch.dtype,
        num_experts: int,
        ep_world_size: int,
        barriers: List[torch.Tensor],
        n_split: int = 4,
        do_all_reduce: bool = False,
        use_read_mode: bool = False,
    ): ...
    def run(
        self,
        group_gemm_buffers: List[torch.Tensor],
        output: Optional[torch.Tensor],
        ep_start: int,
        ep_nexperts: int,
        splits: torch.Tensor,
        routing_idx: torch.Tensor,
        output_vec_scales: List[torch.Tensor],
        num_thread_blocks: int,
        stream: torch.cuda.Stream,
    ) -> torch.Tensor: ...
    def reset_buffer(self) -> None: ...

def prepare_moe_ag_scatter_args(
    splits_gpu: torch.Tensor,
    scatter_index: torch.Tensor,
    ntokens: int,
    topk: int,
    num_weight_groups: int,
    ep_start: int,
    ep_experts: int,
    rank: int,
    world_size: int,
    tile_size_m: int,
    cp_stream: torch.cuda.Stream,
) -> Tuple[
    int,  # M_this_ep, int
    torch.Tensor,  # M_this_ep_pad, torch.Tensor(int32) on device
    torch.Tensor,  # gather_A_index
    torch.Tensor,  # scatter_D_index
    torch.Tensor,  # expert_index
    torch.Tensor,  # rank_start_index
    torch.Tensor,  # rank_end_index
]: ...

class AllGatherOp:
    def __init__(
        self,
        tp_group: dist.ProcessGroup,
        nnodes: int,
        max_m_dim: int,
        k_dim: int,
        input_dtype: torch.dtype,
    ) -> None: ...
    def run(
        self,
        input: torch.Tensor,
        input_scale: Optional[torch.Tensor],
        all_gather_option: AllGatherOption,
        cp_stream: ctypes.c_void_p,
    ) -> None: ...
    def local_barrier_buffer(self) -> torch.Tensor: ...
    def local_input_buffer(self) -> torch.Tensor: ...
    def local_input_scale_buffer(self) -> torch.Tensor: ...

def calc_scatter_index(
    choosed_experts: torch.Tensor,
    splits_gpu: torch.Tensor,
    num_expert: int,
) -> torch.Tensor: ...



################################################################################
#
# Copyright 2025 ByteDance Ltd. and/or its affiliates. All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
################################################################################

import torch
import torch.distributed

import flux
from flux.testing import all_gather_into_tensor_with_fp8, gen_moe_gating_args, matmul_int8
from flux.testing.utils import generate_data


class MoeMlp1Ctx:
    def __init__(
        self,
        TP_GROUP,
        EP_GROUP,
        b: int,
        s: int,
        h: int,  # K
        ffn_size: int,
        nexperts: int,
        topk: int,
        input_dtype: torch.dtype,
        output_dtype: torch.dtype,
        dist: str,
        fast_accum: bool,
        weight_groups: int,
        drop_token: bool,
        debug: bool = False,
        generator: torch.Generator = None,  # torch random generator
        stable: bool = True,
    ) -> None:
        self.b = b
        self.s = s
        self.h = h
        self.ffn_size = ffn_size
        self.nexperts = nexperts
        self.topk = topk
        self.ntokens = b * s
        self.fast_accum = fast_accum
        self.weight_groups = weight_groups
        self.tp_rank = TP_GROUP.rank()
        self.tp_size = TP_GROUP.size()
        self.ep_rank = EP_GROUP.rank()
        self.ep_size = EP_GROUP.size()
        self.ffn_tp_size = self.tp_size // self.ep_size
        self.nexperts_ep = self.nexperts // self.ep_size
        self.drop_token = drop_token
        assert self.nexperts % self.ep_size == 0

        assert self.ffn_size % self.ffn_tp_size == 0
        self.ffn_size_shard = ffn_size // self.ffn_tp_size
        assert self.ntokens % self.tp_size == 0
        self.ntokens_shard = self.ntokens // self.tp_size

        device = torch.cuda.current_device()

        is_s8_dequant = input_dtype == torch.int8
        is_fp8 = flux.is_fp8_dtype(input_dtype)

        if dist == "uniform":
            weights = None
        elif dist == "random_uniform":
            weights = torch.ones(self.nexperts, device=device, dtype=torch.float32)
        else:
            weights = torch.rand(self.nexperts, device=device, dtype=torch.float32)
            if dist == "random_with_first_k_experts":
                weights[self.topk :].fill_(0)

        moe_gating_args = gen_moe_gating_args(
            nexperts,
            topk,
            self.ntokens,
            0.1 if drop_token else 0.0,
            stable=stable,
            weights=weights,
            generator=generator,
        )

        self.splits_gpu = moe_gating_args.splits_gpu
        self.splits_cpu = moe_gating_args.splits_gpu.to("cpu")
        self.scatter_index = moe_gating_args.scatter_index
        self.gather_index = moe_gating_args.gather_index

        self.nrows_ep = int(
            torch.sum(
                self.splits_cpu[
                    self.nexperts_ep * self.ep_rank : self.nexperts_ep * (self.ep_rank + 1)
                ]
            )
        )

        N, K = self.ffn_size_shard, h
        E_this_ep = self.nexperts_ep

        if is_s8_dequant:
            scale_value = 127
        elif is_fp8:
            scale_value = 0.1
        else:
            scale_value = 0.01 * (self.tp_rank + 1)

        data_config = [
            ((self.ntokens_shard, K), input_dtype, (scale_value, 0)),  # input_shard
            ((self.ntokens, K), input_dtype, (1, 0)),  # input_full
            ((self.ntokens * topk, K), input_dtype, (1, 0)),  # scatter_inputs
        ]
        self.inputs_shard, self.inputs, self.scatter_inputs = next(generate_data(data_config))

        data_config = [
            ((E_this_ep, N, K), input_dtype, (scale_value, 0)),  # weights
            ((self.nrows_ep, N), output_dtype, (0, 0)),  # outputs
            ((E_this_ep,), torch.float32, (1.0, 0)),  # output_scale
        ]
        generator = generate_data(data_config)
        self.weights, self.outputs, self.output_scale = [
            *zip(*[list(next(generator)) for _ in range(weight_groups)])
        ]
        if is_s8_dequant:
            data_config = [
                ((E_this_ep, 1, N), torch.float32, (1.0, 0)),  # bias
                ((self.ntokens_shard), torch.float32, (1.0, 0)),  # input_scale_shard
                ((E_this_ep, 1, N), torch.float32, (1.0, 0)),  # weight_scale
                ((self.ntokens, 1), torch.float32, (1.0, 0)),  # input_scale
                ((self.ntokens * topk, 1), torch.float32, (1.0, 0)),  # scatter_input_scale
            ]
            generator = generate_data(data_config)
            (
                self.bias,
                self.input_scale,
                self.weight_scale,
                self.full_input_scale,
                self.scatter_input_scale,
            ) = [*zip(*[list(next(generator)) for _ in range(weight_groups)])]
        else:
            (
                self.bias,
                self.input_scale,
                self.weight_scale,
                self.full_input_scale,
                self.scatter_input_scale,
            ) = (None, None, None, None, None)

        if debug:
            self._fill_debug()

        torch.cuda.synchronize()
        self.is_s8 = is_s8_dequant
        self.is_fp8 = is_fp8

    def _fill_debug(self):
        input_dtype = self.inputs_shard.dtype
        is_s8_dequant = input_dtype == torch.int8
        is_fp8 = flux.is_fp8_dtype(input_dtype)
        rank = self.tp_rank
        with flux.with_torch_deterministic(not is_fp8):
            self.inputs_shard.zero_()
            self.inputs_shard[:, 0] = (
                torch.arange(rank * self.ntokens_shard + 1, (rank + 1) * self.ntokens_shard + 1)
                / 1e2
            )
            # self.inputs_shard.fill_(1 / 32)
            E, _, _ = self.weights[0].shape
            for w in self.weights:
                for e in range(E):
                    w[e, :, :].fill_(e + 1)
            [x.fill_(1) for x in self.output_scale]
            if is_s8_dequant:
                self.bias[0].fill_(0)
                self.input_scale[0].fill_(1)
                self.weight_scale[0].fill_(1)

    def clear_outputs(self):
        for i in range(self.weight_groups):
            self.outputs[i].fill_(0.0)

    def get_outputs_clone(self):
        return [out.clone() for out in self.outputs]


class MoeAgScatterWithTorch(object):
    @staticmethod
    def comm_impl(ctx, TP_GROUP):
        all_gather_into_tensor_with_fp8(ctx.inputs, ctx.inputs_shard, group=TP_GROUP)
        if ctx.is_s8:
            if ctx.input_scale is not None:
                for i in range(ctx.weight_groups):
                    torch.distributed.all_gather_into_tensor(
                        ctx.full_input_scale[i], ctx.input_scale[i], group=TP_GROUP
                    )

    @staticmethod
    def scatter_impl(ctx):
        if flux.is_fp8_dtype(ctx.inputs.dtype):
            ctx.scatter_inputs.view(torch.uint8).copy_(
                torch.index_select(ctx.inputs.view(torch.uint8), dim=0, index=ctx.gather_index)
            )
        else:
            ctx.scatter_inputs.copy_(torch.index_select(ctx.inputs, dim=0, index=ctx.gather_index))

        if ctx.is_s8 and ctx.input_scale is not None:
            for n in range(ctx.weight_groups):
                ctx.scatter_input_scale[n].copy_(
                    torch.index_select(ctx.full_input_scale[n], dim=0, index=ctx.gather_index)
                )

    @staticmethod
    def gemm_impl(ctx, gemm_only_op):
        offset = 0
        expert_id_offset = ctx.nexperts_ep * ctx.ep_rank
        input_offset = torch.sum(ctx.splits_cpu[:expert_id_offset])
        for exp_id in range(ctx.nexperts_ep):
            nxt_offset = offset + ctx.splits_cpu[exp_id + expert_id_offset]
            if nxt_offset - offset > 0:
                exp_input = ctx.scatter_inputs[offset + input_offset : nxt_offset + input_offset]
                for i in range(ctx.weight_groups):
                    out = ctx.outputs[i][offset:nxt_offset]
                    if flux.is_fp8_dtype(exp_input.dtype):
                        gemm_only_op.forward(
                            exp_input,
                            ctx.weights[i][exp_id],
                            output_buf=out,
                            fast_accum=ctx.fast_accum,
                        )
                    elif ctx.is_s8:
                        input_scale = ctx.scatter_input_scale[i][
                            offset + input_offset : nxt_offset + input_offset, :
                        ]
                        accum = matmul_int8(exp_input, ctx.weights[i][exp_id].t())
                        out.copy_(
                            (
                                input_scale * ctx.weight_scale[i][exp_id] * accum.to(torch.float32)
                            ).to(torch.bfloat16)
                        )
                    else:
                        torch.matmul(
                            exp_input,
                            ctx.weights[i][exp_id].t(),
                            out=out,
                        )

                    # TODO(houqi.1993) xperf_gpt has no output_scale. so don't know the order of output_scale and bias, leave it as bias first.
                    if ctx.bias is not None:
                        out.add_(ctx.bias[i][exp_id])
                    if not ctx.is_s8:
                        out.mul_(ctx.output_scale[i][exp_id])

            offset = nxt_offset



################################################################################
#
# Copyright 2025 ByteDance Ltd. and/or its affiliates. All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
################################################################################
__version__ = "1.1.1"
from .cpp_mod import *

if not isinstance(cpp_mod.AGRingMode, cpp_mod.NotCompiled):
    from .ag_kernel_crossnode import *

from .gemm_rs_sm80 import *
from .util import *
from .dist_utils import *



################################################################################
#
# Copyright 2025 ByteDance Ltd. and/or its affiliates. All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
################################################################################

import ctypes
import importlib
import logging
from pathlib import Path

import torch

FLUX_TORCH_EXTENSION_NAME = "flux_ths_pybind"


def _preload_libs(libname):
    libpath = Path(__file__).parent / "lib" / libname
    try:
        ctypes.CDLL(libpath)
    except OSError as e:
        # Try to load from the LD_LIBRARY_PATH
        logging.debug(f"failed to load {libpath}:\n {e}")
        ctypes.CDLL(libname)


def _load_deps():
    try:
        _preload_libs("libnvshmem_host.so.3")
        _preload_libs("nvshmem_bootstrap_uid.so")
        _preload_libs("nvshmem_transport_ibrc.so.3")
    except Exception as e:
        logging.warning("Failed to load NVSHMEM libs")
    _preload_libs("libflux_cuda.so")
    _preload_libs("libflux_cuda_ths_op.so")


_load_deps()
flux_mod = importlib.import_module(FLUX_TORCH_EXTENSION_NAME)


class NotCompiled:
    pass


def _get_flux_member(member):
    return getattr(flux_mod, member, NotCompiled())


def _get_from_torch_classes(name: str):
    try:
        return getattr(torch.classes.flux, name)
    except:
        return NotCompiled()


bsr_reduce = _get_flux_member("bsr_reduce")
init_flux_shm = flux_mod.init_flux_shm
bitwise_check = flux_mod.bitwise_check
uniform_initialize = flux_mod.uniform_initialize
load_tuning_record = flux_mod.load_tuning_record
create_tensor_list = flux_mod.flux_create_tensor_list
GroupBarrier = flux_mod.GroupBarrier

calc_scatter_index = _get_flux_member("calc_scatter_index")

ProfilingContext = flux_mod.ProfilingContext
TuningRecord = flux_mod.TuningRecord
DistEnvTP = flux_mod.DistEnvTP
DistEnvTPWithEP = flux_mod.DistEnvTPWithEP
MoeArguments = flux_mod.MoeArguments

# GEMM only
GemmOnly = _get_flux_member("GemmOnly")
BlockScaleGemm = _get_flux_member("BlockScaleGemm")
GemmGroupedV2 = _get_flux_member("GemmGroupedV2")
GemmGroupedV3 = _get_flux_member("GemmGroupedV3")

# GEMM+RS
GemmRS = _get_flux_member("GemmRS")
get_gemm_rs_threadblock_segments_info = _get_flux_member("get_gemm_rs_threadblock_segments_info")
calc_gemm_rs_threadblock_segments_info = _get_flux_member("calc_gemm_rs_threadblock_segments_info")
GemmRSAcrossNode = _get_flux_member("GemmRSAcrossNode")

# allgather op
AGRingMode = _get_flux_member("AGRingMode")
AllGatherOption = _get_flux_member("AllGatherOption")
if not isinstance(AllGatherOption, NotCompiled):
    AllGatherOption.__repr__ = (
        lambda x: f"AllGatherOption(mode={x.mode}, fuse_sync={x.fuse_sync}, use_cuda_core_local={x.use_cuda_core_local}, use_cuda_core_ag={x.use_cuda_core_ag}, input_buffer_copied={x.input_buffer_copied}, use_read={x.use_read})"
    )
AllGatherOp = _get_flux_member("AllGatherOp")
RingMode = _get_flux_member("RingMode")
ReduceScatterOption = _get_flux_member("ReduceScatterOption")
if not isinstance(ReduceScatterOption, NotCompiled):
    ReduceScatterOption.__repr__ = (
        lambda x: f"use_barrier_queue={x.use_barrier_queue}, "
        "use_1d_ring={x.use_1d_ring}, "
        "use_p2p_read={x.use_p2p_read}, "
        "use_cudaMemcpyAsync={x.use_cudaMemcpyAsync}, "
        "use_gemmk={x.use_gemmk}, "
        "per_tile_flags={x.per_tile_flags}, "
        "n_split={x.n_split}, "
        "num_blocks={x.num_blocks}, "
        "ring_mode={x.ring_mode}"
    )

# AG+GEMM
AGKernel = _get_flux_member("AGKernel")
AGKernelCrossNode = _get_flux_member("AGKernelCrossNode")

# MOE ag-scatter
GemmGroupedV2AGScatterOp = _get_flux_member("GemmGroupedV2AGScatterOp")
GemmGroupedV3AGScatter = _get_flux_member("GemmGroupedV3AGScatter")
prepare_moe_ag_scatter_args = _get_flux_member("prepare_moe_ag_scatter_args")

# MOE gather-rs
GemmGroupedV2GatherRSOp = _get_flux_member("GemmGroupedV2GatherRSOp")
TopkReduceScatterOp = _get_flux_member("TopkReduceScatterOp")
GemmGroupedV3GatherRS = _get_flux_member("GemmGroupedV3GatherRS")
topk_scatter_reduce = _get_flux_member("topk_scatter_reduce")
All2AllOp = _get_flux_member("All2AllOp")

__all__ = [
    "bsr_reduce",
    "init_flux_shm",
    "create_tensor_list",
    "GroupBarrier",
    "bitwise_check",
    "uniform_initialize",
    "topk_scatter_reduce",
    "load_tuning_record",
    "TuningRecord",
    "ProfilingContext",
    "DistEnvTP",
    "DistEnvTPWithEP",
    "MoeArguments",
    "ReduceScatterOption",
    "RingMode",
    "GemmRS",
    "GemmRSAcrossNode",
    "AGKernel",
    "AGKernelCrossNode",
    "All2AllOp",
    "GemmOnly",
    "BlockScaleGemm",
    "GemmGroupedV2",
    "GemmGroupedV3",
    "GemmGroupedV3AGScatter",
    "GemmGroupedV3GatherRS",
    "GemmGroupedV2AGScatterOp",
    "prepare_moe_ag_scatter_args",
    "GemmGroupedV2GatherRSOp",
    "TopkReduceScatterOp",
    "AGRingMode",
    "AllGatherOption",
    "AllGatherOp",
    "get_gemm_rs_threadblock_segments_info",
    "calc_gemm_rs_threadblock_segments_info",
    "calc_scatter_index",
]



################################################################################
#
# Copyright 2025 ByteDance Ltd. and/or its affiliates. All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
################################################################################

from typing import Optional

import torch
import torch.distributed as dist

import flux

from . import cpp_mod

MAX_NUM_SIGNAL = 64
INTER_NODE_AG_KERNEL_GROUP = None
INTRA_NODE_AG_KERNEL_GROUP = None


def get_inter_node_ag_group(tp_group: dist.ProcessGroup, nnodes: int):
    global INTER_NODE_AG_KERNEL_GROUP
    if INTER_NODE_AG_KERNEL_GROUP is not None:
        return INTER_NODE_AG_KERNEL_GROUP
    tp_world_size: int = tp_group.size()
    assert tp_world_size % nnodes == 0, f"{tp_world_size} is not divisible by {nnodes}"

    local_world_size: int = tp_group.size() // nnodes
    world_size: int = dist.get_world_size()
    assert world_size % tp_world_size == 0, f"{world_size} not divisible by {tp_world_size}"

    ranks_per_sub_group = []
    for tp0_rank in range(0, world_size, tp_world_size):
        for local_rank in range(local_world_size):
            inter_ranks = [
                tp0_rank + node_rank * local_world_size + local_rank for node_rank in range(nnodes)
            ]
            ranks_per_sub_group.append(inter_ranks)

    INTER_NODE_AG_KERNEL_GROUP, _ = dist.new_subgroups_by_enumeration(ranks_per_sub_group)
    return INTER_NODE_AG_KERNEL_GROUP


def get_intra_node_ag_group(tp_group: dist.ProcessGroup, nnodes: int):
    global INTRA_NODE_AG_KERNEL_GROUP
    if INTRA_NODE_AG_KERNEL_GROUP is not None:
        return INTRA_NODE_AG_KERNEL_GROUP

    tp_world_size: int = tp_group.size()
    assert tp_world_size % nnodes == 0, f"{tp_world_size} is not divisible by {nnodes}"
    local_world_size: int = tp_group.size() // nnodes
    world_size: int = dist.get_world_size()
    assert world_size % tp_world_size == 0, f"{world_size} not divisible by {tp_world_size}"

    ranks_per_sub_group = []
    for node_id in range(nnodes):
        ranks_per_sub_group.append(
            list(range(node_id * local_world_size, (node_id + 1) * local_world_size))
        )

    INTRA_NODE_AG_KERNEL_GROUP, _ = dist.new_subgroups_by_enumeration(ranks_per_sub_group)
    return INTRA_NODE_AG_KERNEL_GROUP


class AGKernelXNode:
    def __init__(
        self,
        tp_group: dist.ProcessGroup,
        nnodes: int,
        full_m: int,
        n_dim: int,
        k_dim: int,
        input_dtype: torch.dtype,
        transpose_weight: bool = True,
        gather_output: bool = False,
        local_copy: bool = False,
        ring_mode: Optional[cpp_mod.AGRingMode] = None,
    ) -> None:
        """
        : ring_mode[int]
            0: nvlink mode: all-to-all
            1: 1d ring. for PCI-e communication optimization
            2: 2d ring. for PCI-e communication optimization. better performance than 1d ring
        """
        self.tp_group = tp_group
        self.world_size = self.tp_group.size()
        self.local_world_size = self.world_size // nnodes
        self.rank = self.tp_group.rank()
        self.local_rank = self.rank % self.local_world_size
        self.nnodes = nnodes

        self.cross_node = self.nnodes > 1
        if self.cross_node:
            self.inter_node_group = get_inter_node_ag_group(tp_group, nnodes)
        self.intra_node_group = get_intra_node_ag_group(tp_group, nnodes)

        self.input_dtype = input_dtype
        self.full_m = full_m
        self.n_dim = n_dim
        self.k_dim = k_dim
        self.transpose_weight = transpose_weight
        self.gather_output = gather_output
        self.local_copy = local_copy
        # input buffer
        self.input_buffer = None
        # output buffer
        self.output_buffer = None

        self.cpp_op = None
        self.ring_mode = ring_mode
        self.lazy_init_cpp_op(self.full_m)

    def lazy_init_cpp_op(self, full_m: int):
        if self.cpp_op is not None:
            return

        self.full_m = full_m

        # output buffer
        self.output_buffer = torch.zeros(
            (full_m, self.n_dim),
            dtype=self.input_dtype,
            device=torch.cuda.current_device(),
            requires_grad=False,
        )

        self.cpp_op = cpp_mod.AGKernelCrossNode(
            self.tp_group,
            self.intra_node_group,
            self.nnodes,
            self.output_buffer,
            self.full_m,
            self.n_dim,
            self.k_dim,
            self.input_dtype,
            self.transpose_weight,
            self.local_copy,
            self.ring_mode,
        )

    def forward(self, input: torch.Tensor, weight: torch.Tensor) -> torch.Tensor:
        self.lazy_init_cpp_op(self.full_m)
        output = self.cpp_op.forward(input, weight)
        gather_output = self.input_buffer if self.gather_output else None
        return output, gather_output

    def reset_signals(self):
        self.cpp_op.reset_signals()

    def copy_local(self, input: torch.Tensor):
        self.cpp_op.copy_local(input)

    def gemm_only(
        self, input: torch.Tensor, full_input: torch.Tensor, weight: torch.Tensor
    ) -> torch.Tensor:
        output = self.cpp_op.gemm_only(input, full_input, weight)
        return output


__all__ = ["AGKernelXNode"]



import glob
import os
import shutil
import sys
import re
import ast
from pathlib import Path
import urllib
import urllib.request
import urllib.error
import setuptools
import torch
import subprocess
from torch.utils.cpp_extension import BuildExtension
from packaging.version import parse, Version
from typing import Optional, Tuple
from wheel.bdist_wheel import bdist_wheel as _bdist_wheel

# Project directory root
root_path: Path = Path(__file__).resolve().parent
enable_nvshmem = int(os.getenv("FLUX_SHM_USE_NVSHMEM", 0))
PACKAGE_NAME = "byte_flux"
BASE_WHEEL_URL = "https://github.com/bytedance/flux/releases/download/{tag_name}/{wheel_name}"
FORCE_BUILD = os.getenv("FLASH_ATTENTION_FORCE_BUILD", "FALSE") == "TRUE"
USE_LOCAL_VERSION = int(os.getenv("FLUX_USE_LOCAL_VERSION", 0))


def cuda_version() -> Tuple[int, ...]:
    """CUDA Toolkit version as a (major, minor) by nvcc --version"""

    # Try finding NVCC
    nvcc_bin: Optional[Path] = None
    if nvcc_bin is None and os.getenv("CUDA_HOME"):
        # Check in CUDA_HOME
        cuda_home = Path(os.getenv("CUDA_HOME"))
        nvcc_bin = cuda_home / "bin" / "nvcc"
    if nvcc_bin is None:
        # Check if nvcc is in path
        nvcc_bin = shutil.which("nvcc")
        if nvcc_bin is not None:
            nvcc_bin = Path(nvcc_bin)
    if nvcc_bin is None:
        # Last-ditch guess in /usr/local/cuda
        cuda_home = Path("/usr/local/cuda")
        nvcc_bin = cuda_home / "bin" / "nvcc"
    if not nvcc_bin.is_file():
        raise FileNotFoundError(f"Could not find NVCC at {nvcc_bin}")

    # Query NVCC for version info
    output = subprocess.run(
        [nvcc_bin, "-V"],
        capture_output=True,
        check=True,
        universal_newlines=True,
    )
    match = re.search(r"release\s*([\d.]+)", output.stdout)
    version = match.group(1).split(".")
    return tuple(int(v) for v in version)


def get_local_version(public_version):
    cuda_version_major, cuda_version_minor = cuda_version()
    torch_version_splits = torch.__version__.split(".")
    torch_version = f"{torch_version_splits[0]}.{torch_version_splits[1]}"
    version = (
        public_version + f"+cu{cuda_version_major}{cuda_version_minor}" + f"torch{torch_version}"
    )
    return version


def get_public_version():
    with open(Path(root_path) / "python" / "flux" / "__init__.py", "r") as f:
        version_match = re.search(r"^__version__\s*=\s*(.*)$", f.read(), re.MULTILINE)
    public_version = ast.literal_eval(version_match.group(1))
    return public_version


def get_package_version():
    global USE_LOCAL_VERSION
    public_version = get_public_version()
    if USE_LOCAL_VERSION:
        return get_local_version(public_version)
    return public_version


def pathlib_wrapper(func):
    def wrapper(*kargs, **kwargs):
        include_dirs, library_dirs, libraries = func(*kargs, **kwargs)
        return map(str, include_dirs), map(str, library_dirs), map(str, libraries)

    return wrapper


@pathlib_wrapper
def cutlass_deps():
    cutlass_home = root_path / "3rdparty/cutlass"
    include_dirs = [
        cutlass_home / "include",
        cutlass_home / "tools" / "util" / "include",
        cutlass_home / "tools" / "library" / "include",
        cutlass_home / "tools" / "profiler" / "include",
    ]
    library_dirs = []
    libraries = []
    return include_dirs, library_dirs, libraries


def read_flux_ths_targets():
    file_path = root_path / "build/src/ths_op/flux_ths_targets.txt"
    variables = {}
    if not os.path.exists(file_path):
        # flux is installed through pip3, the flux_ths_files.txt is not generated
        return []
    with open(file_path, "r") as file:
        for line in file:
            if "=" in line:
                key, value = line.strip().split("=", 1)
                variables[key] = value
    return [x for x in variables["FLUX_THS_TARGETS"].split(";") if x]


@pathlib_wrapper
def nvshmem_deps():
    nvshmem_home = Path(os.environ.get("NVSHMEM_HOME", root_path / "3rdparty/nvshmem/build/src"))
    include_dirs = [nvshmem_home / "include"]
    library_dirs = [nvshmem_home / "lib"]
    # libraries = ["nvshmem"]
    libraries = ["nvshmem_host"]
    return include_dirs, library_dirs, libraries


@pathlib_wrapper
def flux_cuda_deps():
    include_dirs = [root_path / "include", root_path / "src"]
    library_dirs = [root_path / "build" / "lib"]
    libraries = ["flux_cuda", "flux_cuda_ths_op"]
    return include_dirs, library_dirs, libraries


@pathlib_wrapper
def cuda_deps():
    cuda_home = Path(os.environ.get("CUDA_HOME", "/usr/local/cuda"))
    include_dirs = [cuda_home / "include"]
    library_dirs = [cuda_home / "lib64", cuda_home / "lib64/stubs"]
    libraries = ["cuda", "cudart", "nvidia-ml"]
    return include_dirs, library_dirs, libraries


@pathlib_wrapper
def nccl_deps():
    nccl_home = Path(os.environ.get("NCCL_ROOT", root_path / "3rdparty/nccl/build/local"))
    include_dirs = [nccl_home / "include", nccl_home / "include" / "nccl" / "detail" / "include"]
    library_dirs = [nccl_home / "lib"]
    libraries = ["nccl_static"]
    return include_dirs, library_dirs, libraries


def setup_pytorch_extension() -> setuptools.Extension:
    """Setup CppExtension for PyTorch support"""
    include_dirs, library_dirs, libraries = [], [], []

    deps = [nccl_deps(), cutlass_deps(), flux_cuda_deps(), cuda_deps()]
    if enable_nvshmem:
        deps.append(nvshmem_deps())
    for include_dir, library_dir, library in deps:
        include_dirs += include_dir
        library_dirs += library_dir
        libraries += library

    # Compiler flags
    # too much warning from CUDA /usr/local/cuda/include/cusparse.h: "-Wdeprecated-declarations"
    cxx_flags = [
        "-O3",
        "-DTORCH_CUDA=1",
        "-fvisibility=hidden",
        "-Wno-deprecated-declarations",
        "-fdiagnostics-color=always",
    ]
    if enable_nvshmem:
        cxx_flags.append("-DFLUX_SHM_USE_NVSHMEM")
    ld_flags = ["-Wl,--exclude-libs=libnccl_static"]
    flux_ths_targets = [
        str(x.relative_to(root_path))  # relative path for include_package_data
        for x in Path(root_path / "src" / "pybind").glob("*.cc")
        if x.stem in read_flux_ths_targets()
    ]

    from torch.utils.cpp_extension import CppExtension

    return CppExtension(
        name="flux_ths_pybind",
        sources=flux_ths_targets,
        include_dirs=include_dirs,
        library_dirs=library_dirs,
        libraries=libraries,
        extra_compile_args=cxx_flags,
        extra_link_args=ld_flags,
    )

def get_wheel_url():
    flux_tag_version = get_public_version()
    python_version = f"cp{sys.version_info.major}{sys.version_info.minor}"
    torch_version_raw = parse(torch.__version__)
    torch_version = f"{torch_version_raw.major}.{torch_version_raw.minor}"
    torch_cuda_version = parse(torch.version.cuda)
    torch_cuda_version = parse("11.8") if torch_cuda_version.major == 11 else parse("12.4")
    cuda_version = f"{torch_cuda_version.major}{torch_cuda_version.minor}"
    wheel_filename = f"{PACKAGE_NAME}-{flux_tag_version}+cu{cuda_version}torch{torch_version}-{python_version}-{python_version}-linux_x86_64.whl"
    wheel_url = BASE_WHEEL_URL.format(tag_name=f"v{flux_tag_version}", wheel_name=wheel_filename)
    return wheel_url, wheel_filename

class CachedWheelsCommand(_bdist_wheel):
    """
    The CachedWheelsCommand plugs into the default bdist wheel, which is ran by pip when it cannot
    find an existing wheel (which is currently the case for all flash attention installs). We use
    the environment parameters to detect whether there is already a pre-built version of a compatible
    wheel available and short-circuits the standard full build pipeline.
    """

    def run(self):
        if FORCE_BUILD:
            return super().run()

        wheel_url, wheel_filename = get_wheel_url()
        try:
            urllib.request.urlretrieve(wheel_url, wheel_filename)

            # Make the archive
            # Lifted from the root wheel processing command
            # https://github.com/pypa/wheel/blob/cf71108ff9f6ffc36978069acb28824b44ae028e/src/wheel/bdist_wheel.py#LL381C9-L381C85
            if not os.path.exists(self.dist_dir):
                os.makedirs(self.dist_dir)

            impl_tag, abi_tag, plat_tag = self.get_tag()
            archive_basename = f"{self.wheel_dist_name}-{impl_tag}-{abi_tag}-{plat_tag}"

            wheel_path = os.path.join(self.dist_dir, archive_basename + ".whl")
            print("Raw wheel path", wheel_path)
            shutil.move(wheel_filename, wheel_path)
        except (urllib.error.HTTPError, urllib.error.URLError):
            print("Precompiled wheel not found. Building from source...")
            # If the wheel could not be downloaded, build from source
            super().run()

def main():
    flux_version = get_package_version()
    packages = setuptools.find_packages(
        where="python",
        include=[
            "flux",
            "flux.testing",
        ],
    )
    data_file_list = ["python/flux/lib/libflux_cuda.so"]
    data_file_list += ["python/flux/lib/libflux_cuda_ths_op.so"]
    if enable_nvshmem:
        data_file_list += [
            "python/flux/lib/nvshmem_bootstrap_uid.so",
            "python/flux/lib/nvshmem_transport_ibrc.so.3",
            "python/flux/lib/libnvshmem_host.so.3",
        ]
    # Configure package
    setuptools.setup(
        name=PACKAGE_NAME,
        version=flux_version,
        package_dir={"": "python"},
        packages=packages,
        description="Flux library",
        ext_modules=[setup_pytorch_extension()],
        cmdclass={"bdist_wheel": CachedWheelsCommand, "build_ext": BuildExtension},
        setup_requires=["torch", "cmake", "packaging"],
        install_requires=["torch"],
        extras_require={"test": ["torch", "numpy"]},
        license_files=("LICENSE",),
        package_data={
            "python/flux/lib": ["*.so"],
            "python/flux/include": ["*.h"],
            "python/flux/share": ["*.cmake"],
        },  # only works for bdist_wheel under package
        python_requires=">=3.8",
        include_package_data=True,
        data_files=[
            (
                "lib",  # installed directory
                data_file_list,  # to installed shared libraries. only works for setup.py install/bdist_wheels
            )
        ],
    )


if __name__ == "__main__":
    main()



################################################################################
#
# Copyright 2025 ByteDance Ltd. and/or its affiliates. All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
################################################################################

import datetime
import os
import random
from typing import Callable

import numpy as np
import torch
import torch.distributed


class DistEnv:
    def __init__(self, deterministic: bool = True) -> None:
        self.RANK = int(os.environ.get("RANK", 0))
        self.LOCAL_RANK = int(os.environ.get("LOCAL_RANK", 0))
        self.LOCAL_WORLD_SIZE = int(os.environ.get("LOCAL_WORLD_SIZE", 1))
        self.WORLD_SIZE = int(os.environ.get("WORLD_SIZE", 1))
        self.NNODES = self.WORLD_SIZE // self.LOCAL_WORLD_SIZE

        self.init_global_group()
        if deterministic:
            self.setup_deterministic(self.RANK)

    def setup_deterministic(self, init_seed: int) -> None:
        os.environ["NCCL_DEBUG"] = "ERROR"
        os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":16:8"
        torch.use_deterministic_algorithms(True, warn_only=True)
        torch.set_printoptions(precision=8)
        seed = init_seed + self.RANK
        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        torch.backends.cuda.matmul.allow_tf32 = False
        torch.backends.cuda.matmul.allow_fp16_reduced_precision_reduction = False
        torch.backends.cuda.matmul.allow_bf16_reduced_precision_reduction = False
        np.random.seed(seed)
        random.seed(seed)

    def init_global_group(self) -> None:
        torch.distributed.init_process_group(
            backend="nccl",
            world_size=self.WORLD_SIZE,
            rank=self.RANK,
            timeout=datetime.timedelta(seconds=1800),
        )

    def get_world(self):
        return torch.distributed.group.WORLD

    def new_group(self, ranks):
        return torch.distributed.new_group(ranks=ranks, backend="nccl")


DIST_ENV = None


def get_dist_env(deterministic: bool = True):
    global DIST_ENV

    if DIST_ENV == None:
        DIST_ENV = DistEnv(deterministic=deterministic)
    return DIST_ENV


def exec_in_rank_order(group: torch.distributed.ProcessGroup, func: Callable):
    for i in range(group.size()):
        if i == group.rank():
            func()
        group.barrier()
        torch.cuda.synchronize()


__all__ = ["get_dist_env", "exec_in_rank_order"]



################################################################################
#
# Copyright 2025 ByteDance Ltd. and/or its affiliates. All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
################################################################################

from dataclasses import dataclass
from functools import lru_cache
from typing import Optional

import torch
import torch.distributions

try:
    import triton
    import triton.language as tl
except Exception as e:
    pass

import flux

INDEX_DTYPE = torch.int64

_DISTRIBUTION = {
    "xl-4in32-training": [
        0.025833333333333333,
        0.026041666666666668,
        0.0209375,
        0.02,
        0.051041666666666666,
        0.05364583333333333,
        0.018958333333333334,
        0.0328125,
        0.016979166666666667,
        0.030625,
        0.04822916666666666,
        0.05572916666666667,
        0.019166666666666665,
        0.036458333333333336,
        0.015833333333333335,
        0.0128125,
        0.029583333333333333,
        0.03770833333333334,
        0.04895833333333333,
        0.027916666666666666,
        0.036145833333333335,
        0.03197916666666667,
        0.022083333333333333,
        0.022395833333333334,
        0.0225,
        0.0278125,
        0.021875,
        0.04614583333333333,
        0.03239583333333333,
        0.0384375,
        0.019895833333333335,
        0.0490625,
    ]
}


@lru_cache(maxsize=8)
def _pre_distribution(key, device_index):
    return torch.tensor(_DISTRIBUTION[key], dtype=torch.float32).cuda(device_index)


@dataclass
class moe_gating_args:
    choosed_experts: torch.Tensor
    scatter_index: torch.Tensor
    gather_index: torch.Tensor
    topk_index: torch.Tensor
    splits_gpu: torch.Tensor


def calc_gather_index_stable(choosed_experts: torch.Tensor):
    _, index_choosed_experts = choosed_experts.flatten().sort(stable=True)
    gather_index = index_choosed_experts.to(torch.int32) // topk
    topk_index = torch.arange(0, topk, dtype=torch.int32, device="cuda").repeat(ntokens)[
        index_choosed_experts
    ]
    return gather_index, topk_index


def calc_gather_index(
    scatter_index: torch.Tensor,
    row_start: int,
    row_end: int,
    BLOCK_SIZE: int = 1024,
):
    @triton.jit
    def _kernel(
        scatter_index: torch.Tensor,
        gather_index: torch.Tensor,
        topk_index: torch.Tensor,
        ntokens: int,
        topk: int,
        row_start: int,
        row_end: int,
        BLOCK_SIZE: tl.constexpr,
    ):
        pid = tl.program_id(axis=0)
        offset = pid * BLOCK_SIZE + tl.arange(0, BLOCK_SIZE)
        mask = offset < ntokens * topk
        scatter_idx = tl.load(scatter_index + offset, mask=mask, other=-1)
        token_idx = offset // topk
        topk_idx = offset % topk
        token_idx_mask = (scatter_idx >= row_start) & (scatter_idx < row_end)
        tl.store(gather_index + scatter_idx - row_start, token_idx, mask=token_idx_mask)
        tl.store(topk_index + scatter_idx - row_start, topk_idx, mask=token_idx_mask)

    ntokens, topk = scatter_index.shape
    gather_index = torch.zeros(row_end - row_start, dtype=torch.int32, device=scatter_index.device)
    topk_index = torch.zeros(row_end - row_start, dtype=torch.int32, device=scatter_index.device)
    grid = lambda META: (triton.cdiv(ntokens * topk, META["BLOCK_SIZE"]),)
    _kernel[grid](
        scatter_index,
        gather_index,
        topk_index,
        ntokens,
        topk,
        row_start,
        row_end,
        BLOCK_SIZE=BLOCK_SIZE,
        num_warps=BLOCK_SIZE // 32,
    )
    return gather_index, topk_index


def calc_scatter_index_stable(choosed_experts):
    return (
        choosed_experts.flatten().argsort(stable=True).argsort().int().view(choosed_experts.shape)
    )


def calc_scatter_index_stable_sync(choosed_experts, splits_gpu):
    scatter_index = torch.ones_like(choosed_experts) * -1
    cumsum_gpu = splits_gpu.cumsum(dim=0) - splits_gpu
    E = splits_gpu.numel()
    for e in range(E):
        scatter_index[choosed_experts == e] = torch.arange(
            cumsum_gpu[e], cumsum_gpu[e] + splits_gpu[e], dtype=torch.int32, device="cuda"
        )  # this step is synchronized


@torch.no_grad()
def gen_moe_gating_args(
    E: int,
    topk: int,
    ntokens: int,
    drop_token_ratio: float = 0.0,
    weights: Optional[torch.Tensor] = None,
    generator: torch.Generator = None,
    stable=True,
):
    """
    Args:
        E: number of experts
        topk: number of experts to be selected
        ntokens: number of tokens
        drop_token_ratio: ratio of tokens to be dropped
        weights: weights[eid] * notkens for experts. shape [E].
            NOTE:
                * if weights is not set, use even distribution
                * weights only *guides* the token count by distribution, not precisely set the count
        generator: torch.Generator
        stable: whether to use stable moe gating
    Returns:
        moe_gating_args:
            choosed_experts: shape [ntokens, topk]
            scatter_index: shape [ntokens, topk]
            gather_index: shape [ntokens * topk]
            splits_gpu: shape [E]
            topk_index: shape [ntokens * topk]
    """
    if weights is not None:
        assert len(weights.shape) == 1 and weights.shape[0] == E, f"weights: {weights.shape}"
        if not weights.is_cuda:
            weights = weights.to("cuda")
    drop_token = drop_token_ratio > 0.0
    if weights is None:
        # use the very deterministic moe gating. don't respect drop_ratio.
        if drop_token:
            E = E + 1
        choosed_experts = (
            torch.arange(E, device="cuda", dtype=torch.int32)
            .repeat(ntokens, 1)
            .flatten()[: topk * ntokens]
            .reshape(ntokens, topk)
        )
    else:
        if drop_token:
            assert drop_token_ratio > 0.0 and drop_token_ratio < 1.0
            weights_with_drop = torch.zeros(E + 1, device="cuda", dtype=torch.float32)
            weights_with_drop[:E].copy_(weights)
            weights_with_drop[-1] = weights.sum() / (1 - drop_token_ratio) * drop_token_ratio
            weights = weights_with_drop
            E = E + 1

        choosed_experts = torch.multinomial(
            weights.repeat(ntokens, 1), topk, replacement=False, generator=generator
        ).to(torch.int32)

    splits_gpu = torch.bincount(choosed_experts.view(-1), minlength=E).to(
        torch.int32
    )  # this step is synchronized

    if stable:
        scatter_index = calc_scatter_index_stable(choosed_experts)
    else:
        scatter_index = flux.calc_scatter_index(choosed_experts, splits_gpu)

    try:
        gather_index, topk_index = calc_gather_index(scatter_index, 0, ntokens * topk)
    except Exception as e:
        assert stable == True, f"non-stable gather_index calculation requires triton"
        gather_index, topk_index = calc_gather_index_stable(choosed_experts)

    return moe_gating_args(
        choosed_experts=choosed_experts,
        scatter_index=scatter_index,
        gather_index=gather_index,
        topk_index=topk_index,
        splits_gpu=splits_gpu,
    )


@torch.no_grad()
def gate_func(
    token_num,
    num_experts,
    topk,
    dist: str = "random",
    generator: torch.Generator = None,
):
    if dist == "random":
        weights = torch.rand(num_experts, dtype=torch.float32, device="cuda", generator=generator)
    elif dist == "random_with_first_k_experts":
        weights = torch.rand(num_experts, dtype=torch.float32, device="cuda", generator=generator)
        weights[topk:].fill_(0)
    elif dist == "uniform":
        weights = torch.ones(num_experts, dtype=torch.float32, device="cuda")
    else:
        raise ValueError(f"Unknown distribution: {dist}")

    weights = torch.ones(num_experts, dtype=torch.float32, device="cuda")
    args = gen_moe_gating_args(
        E=num_experts,
        topk=topk,
        ntokens=token_num,
        drop_token_ratio=0.0,
        weights=weights,
        generator=generator,
    )
    return (
        args.splits_gpu.to("cpu"),
        args.choosed_experts,
        args.gather_index,
        args.topk_index,
        args.scatter_index.flatten(),
    )


def _is_strictly_monotone_increasing(tensor: torch.Tensor) -> bool:
    return (tensor[1:] - tensor[:-1] > 0).all()


def is_gather_index_stable(gather_index: torch.Tensor, splits_gpu: torch.Tensor):
    """check if for all experts, gather_index is strictly monotone increasing"""
    # gather_index: [E * topk]
    # splits_gpu: [E]
    splits_cpu = splits_gpu.to("cpu")
    splits_cpu_cumsum = torch.zeros(splits_cpu.shape[0] + 1, dtype=torch.int32, device="cpu")
    splits_cpu_cumsum[1:].copy_(splits_cpu.cumsum(0))
    splits_cpu_cumsum = splits_cpu_cumsum.tolist()
    return all(
        [
            _is_strictly_monotone_increasing(gather_index[s:e])
            for s, e in zip(splits_cpu_cumsum[:-1], splits_cpu_cumsum[1:])
        ]
    )


if __name__ == "__main__":
    topk = 5
    E = 10
    ntokens = 10

    args = gen_moe_gating_args(E, topk, ntokens, drop_token_ratio=0.5)
    print(f"with drop ratio = 0.5")
    print(f"Arguments: {args}")
    print(f" {args.choosed_experts.dtype}")
    print(f" {args.gather_index.dtype}")
    print(f" {args.scatter_index.dtype}")
    print(f" {args.splits_gpu.dtype}")
    print(f" {args.topk_index.dtype}")
    print("\n\n\n")

    weights = torch.arange(E, device="cuda", dtype=torch.float32)
    args = gen_moe_gating_args(E, topk, ntokens, weights=weights)
    print(f"with weights: {weights}")
    print(f"Arguments: {args}")
    print("\n\n\n")

    args = gen_moe_gating_args(E, topk, ntokens)
    print(f"Arguments: {args}")

    ntokens = 10000
    args = gen_moe_gating_args(E, topk, ntokens, stable=False)
    print(f"Arguments: {args}")
    if is_gather_index_stable(args.gather_index, args.splits_gpu):
        print("gather_index is sorted")
    else:
        print("gather_index is not sorted")



################################################################################
#
# Copyright 2025 ByteDance Ltd. and/or its affiliates. All rights reserved.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
################################################################################

# TODO(houqi.1993) some code comes from triton. copy but how to write the copyright?
from typing import Optional
import torch
import subprocess
import sys
import functools
import logging

_SHARED_MEMORY_SIZE = {"NVIDIA L20": 101376}


def get_device_shared_memory_size():
    try:
        name = torch.cuda.get_device_name(torch.cuda.current_device())
        return _SHARED_MEMORY_SIZE[name]
    except Exception:
        from cuda import cudart

        code, shared_memory_size = cudart.cudaDeviceGetAttribute(
            cudart.cudaDeviceAttr.cudaDevAttrMaxSharedMemoryPerMultiprocessor, 0
        )
        assert code == cudart.cudaError_t.cudaSuccess
        return shared_memory_size


def is_fp8_dtype(dtype: torch.dtype):
    return dtype.itemsize == 1 and dtype.is_floating_point


def nvsmi(attrs):
    attrs = ",".join(attrs)
    cmd = ["nvidia-smi", "-i", "0", "--query-gpu=" + attrs, "--format=csv,noheader,nounits"]
    out = subprocess.check_output(cmd)
    ret = out.decode(sys.stdout.encoding).split(",")
    ret = [int(x) for x in ret]
    return ret


@functools.lru_cache()
def get_clock_rate_in_khz():
    try:
        return nvsmi(["clocks.max.sm"])[0] * 1e3
    except FileNotFoundError:
        import pynvml

        pynvml.nvmlInit()
        handle = pynvml.nvmlDeviceGetHandleByIndex(0)
        return pynvml.nvmlDeviceGetMaxClockInfo(handle, pynvml.NVML_CLOCK_SM) * 1e3


@functools.lru_cache()
def get_device_property(device):
    return torch.cuda.get_device_properties(device=device)


@functools.lru_cache()
def get_device_multi_processor_count(device):
    return get_device_property(device).multi_processor_count


def get_max_tensorcore_tflops(dtype, clock_rate, device=None):
    device = device or torch.cuda.current_device()
    num_subcores = get_device_multi_processor_count(device) * 4  # on recent GPUs
    capability = torch.cuda.get_device_capability(device)
    if capability[0] < 8:
        assert dtype == torch.float16
        ops_per_sub_core = 256  # 2 4x4x4 Tensor Cores
    else:
        if dtype in [torch.float32, torch.int32]:
            ops_per_sub_core = 256
        elif dtype in [torch.float16, torch.bfloat16, torch.int16]:
            ops_per_sub_core = 512  # 16x8x32
        elif is_fp8_dtype(dtype) or dtype == torch.int8:
            ops_per_sub_core = 1024  #
        else:
            raise RuntimeError("dtype not supported")
    tflops = num_subcores * clock_rate * ops_per_sub_core * 1e-9
    logging.info(
        f"num_subcores: {num_subcores} clock_rate: {clock_rate} ops_per_sub_core: {ops_per_sub_core}"
    )
    return tflops


def get_max_simd_tflops(dtype, clock_rate, device=None):
    if not device:
        device = torch.cuda.current_device()

    num_subcores = get_device_multi_processor_count(device) * 4
    capability = torch.cuda.get_device_capability()
    if capability[0] < 8:
        if dtype == torch.float32:
            ops_per_sub_core = 32  # 2*16
        elif dtype == torch.float16:
            ops_per_sub_core = 64
        else:
            raise RuntimeError("dtype not supported")
    else:
        if dtype == torch.float32:
            ops_per_sub_core = 32
        elif dtype in [torch.float16, torch.bfloat16]:
            ops_per_sub_core = 64
        else:
            raise RuntimeError("dtype not supported")
    tflops = num_subcores * clock_rate * ops_per_sub_core * 1e-9
    return tflops


def get_tensorcore_tflops(device, num_ctas, num_warps, dtype):
    """return compute throughput in TOPS"""
    total_warps = num_ctas * min(num_warps, 4)
    num_subcores = get_device_multi_processor_count(device) * 4  # on recent GPUs
    tflops = (
        min(num_subcores, total_warps)
        / num_subcores
        * get_max_tensorcore_tflops(dtype, get_clock_rate_in_khz(), device)
    )
    return tflops


def get_simd_tflops(device, num_ctas, num_warps, dtype):
    """return compute throughput in TOPS"""
    total_warps = num_ctas * min(num_warps, 4)
    num_subcores = get_device_multi_processor_count(device) * 4  # on recent GPUs
    tflops = (
        min(num_subcores, total_warps)
        / num_subcores
        * get_max_simd_tflops(dtype, get_clock_rate_in_khz(), device)
    )
    return tflops


def get_tflops_approx(device: torch.dtype, num_ctas: int, num_warps: int, dtype: torch.dtype):
    """You may not achieve"""
    capability = torch.cuda.get_device_capability(device)
    if capability[0] < 8 and dtype == torch.float32:
        return get_simd_tflops(device, num_ctas, num_warps, dtype)
    return get_tensorcore_tflops(device, num_ctas, num_warps, dtype)


def get_full_tflops_approx(dtype: torch.dtype, device: Optional[torch.device] = None):
    prop = torch.cuda.get_device_properties(device)
    return get_tflops_approx(device, prop.multi_processor_count, 4, dtype)


def get_tflops(dtype):
    """TFLOPS with no sparse."""
    device_name = torch.cuda.get_device_name(
        torch.cuda.current_device()
    )  # arch is not a good idea. using device name is better.
    is_fp16 = dtype in [torch.bfloat16, torch.float16]
    is_fp8 = is_fp8_dtype(dtype)
    is_s8 = dtype == torch.int8
    assert is_fp16 or is_fp8 or is_s8
    # https://www.nvidia.com/content/dam/en-zz/Solutions/Data-Center/a100/pdf/nvidia-a100-datasheet-nvidia-us-2188504-web.pdf
    # "NVIDIA A100 80GB PCIe" or "NVIDIA A100-SXM4-80GB" or "A100-SXM4-40GB"
    if device_name.find("A100") >= 0 or device_name.find("A800") >= 0:
        assert is_fp16 or is_s8
        return 312 if is_fp16 else 624
    if device_name == "NVIDIA L20":  # No doc from NVIDIA
        return 119 if is_fp16 else 239
    # https://www.nvidia.com/en-us/data-center/l4/
    if device_name == "NVIDIA L4":
        return 121 if is_fp16 else 242
    # https://images.nvidia.com/content/Solutions/data-center/vgpu-L40-datasheet.pdf
    if device_name == "NVIDIA L40":
        return 181 if is_fp16 else 362
    # https://www.nvidia.com/en-us/data-center/l40s/
    if device_name == "NVIDIA L40S":
        return 366 if is_fp16 else 733
    # https://www.nvidia.com/en-us/data-center/h100/
    if device_name == "NVIDIA H100" or device_name == "NVIDIA H800":
        return 989 if is_fp16 else 1979
    if device_name == "NVIDIA H20":
        return 148 if is_fp16 else 296

    logging.warning(f"device {device_name} not listed in flux. calculate tflops by estimation")

    return get_full_tflops_approx(dtype=dtype)


def get_dram_gbps_by_device_name(device_name: str):
    _DRAM_GBPS = {
        "NVIDIA L20": 864,
        "NVIDIA L4": 300,
        "NVIDIA L40": 864,
        "NVIDIA L40S": 864,
        "NVIDIA H20": 4000,
        "NVIDIA A100 80GB PCIe": 1935,
        "NVIDIA A100-SXM4-80GB": 2039,
        "NVIDIA A100-SXM4-40GB": 1555,
        "NVIDIA H100 SXM": 3958,
        "NVIDIA H100 NVL": 3341,
    }
    return _DRAM_GBPS[device_name]


def get_dram_gbps(device=None):
    try:
        import triton

        return triton.testing.get_dram_gbps(device)
    except:
        return get_dram_gbps_by_device_name(torch.cuda.get_device_name(device))


def estimate_gemm_sol_time_ms(M: int, N: int, K: int, dtype=torch.bfloat16):
    """refer to this: https://arnon.dk/matching-sm-architectures-arch-and-gencode-for-various-nvidia-cards/"""

    flops = M * N * K * 2
    return flops / get_tflops(dtype=dtype) / 1e9


def _get_output_dtype(input_dtype):
    is_s8 = input_dtype == torch.int8
    is_fp8 = is_fp8_dtype(input_dtype)
    return torch.bfloat16 if is_s8 and is_fp8 else input_dtype


def print_gemm_sol_time(M, N, K, input_dtype, output_dtype=None):
    """return time in milliseconds"""
    output_dtype = output_dtype or _get_output_dtype(input_dtype)
    flops = M * K * N * 2
    bytes_r = (M * K + N * K) * input_dtype.itemsize
    bytes_w = (M * N) * output_dtype.itemsize
    t_c = flops / 1e9 / get_tflops(input_dtype)
    t_r, t_w = bytes_r / 1e6 / get_dram_gbps(), bytes_w / 1e6 / get_dram_gbps()
    print(f"SOL gemm TensorCore {t_c:0.3f}ms Memory read: {t_r:0.3f}ms write: {t_w:0.3f}ms")
    return t_c, t_r, t_w


def print_grouped_gemm_sol_time_ms(
    M: int,
    N: int,
    K: int,
    E: int,
    input_dtype: torch.dtype,
    output_dtype: torch.dtype = None,
    num_groups: int = 1,
):
    output_dtype = output_dtype or _get_output_dtype(input_dtype)
    flops = M * K * N * 2 * num_groups
    bytes_r = (M * K + N * K * E) * input_dtype.itemsize * num_groups
    bytes_w = (M * N) * output_dtype.itemsize * num_groups
    t_c = flops / 1e9 / get_tflops(input_dtype)
    t_r, t_w = bytes_r / 1e6 / get_dram_gbps(), bytes_w / 1e6 / get_dram_gbps()
    print(f"SOL grouped gemm TensorCore {t_c:0.3f}ms Memory read: {t_r:0.3f}ms write: {t_w:0.3f}ms")
    return t_c, t_r, t_w


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    for dtype in [
        torch.float16,
        torch.bfloat16,
        torch.float8_e4m3fn,
        torch.float8_e5m2,
        torch.int8,
    ]:
        tflops = get_tflops(dtype)
        tflops_approx = get_full_tflops_approx(dtype)
        print(f"{dtype}: {tflops} TFOLPS or {tflops_approx:0.1f} TFLOPS")
